from itertools import chain
from typing import List, Optional
from pydantic import BaseModel
from vpp_vrouter.client import ExtendedVPPAPIClient
from vpp_vrouter.common import models
from loguru import logger as log
import ipaddress
from vrouter_agent.utils.cli import run_command
from vrouter_agent.utils.config import convert_wg_to_gre_ip_addr
from vrouter_agent.utils.gre_manager import get_gre_manager, convert_wg_to_gre_ip_addr_safe
from vrouter_agent.models import Tunnel
from vrouter_agent.core.base import CreatedWireguardTunnel, Interface
from vrouter_agent.core.config import settings
from vrouter_agent.services.frr import FRR
from vrouter_agent.schemas.tunnel_config import WireGuardConfig

    
class VRouterClient:
    """
    A class to manage VPP API client operations.
    """
    
    def __init__(self, tunnels: List[WireGuardConfig], frr_config:Optional[str]=None, lcp_global: str = "frr"):
        self.client = ExtendedVPPAPIClient()
        self.tunnels = tunnels
        self.wireguard_tunnels: Optional[List[CreatedWireguardTunnel]] = None
        self.gre_tunnels = None
        self.lcp_global = lcp_global
        self.frr_config = frr_config
        
        
    def is_connected(self):
        """
        Check if the client is connected to the vRouter.
        vRouter is the VPP agent.
        Returns:
            bool: True if connected, False otherwise.
        """
        try:
            self.client.get_configuration()
            self.add_lcp_global_configuration()
            return True
        except Exception as e:
            log.error(f"Error connecting to vRouter: {e}")
            return False
    def _get_next_available_wg_index(self, start_index: int = 0) -> int:
        """
        Get the next available WireGuard interface index by checking existing interfaces.
        Args:
            start_index (int): The starting index to check from.
        Returns:
            int: The next available index for WireGuard interface naming.
        """
        try:
            current_config = self.client.get_configuration()
            existing_wg_names = set()
            
            # Extract existing WireGuard interface names
            for item in current_config.items:
                if hasattr(item.config, 'name') and hasattr(item.config, 'type'):
                    if (hasattr(models.InterfaceType, 'WIREGUARD_TUNNEL') and 
                        item.config.type == models.InterfaceType.WIREGUARD_TUNNEL):
                        existing_wg_names.add(item.config.name)
            
            # Find the next available index
            index = start_index
            while f"wg{index}" in existing_wg_names:
                index += 1
                
            log.debug(f"Next available WireGuard index: {index}, existing interfaces: {existing_wg_names}")
            return index
            
        except Exception as e:
            log.warning(f"Could not check existing WireGuard interfaces: {e}, starting from index {start_index}")
            return start_index

    def create_wireguard_tunnels(self) -> dict:
        """
        Create wireguard tunnels.
        Returns:
            dict: Contains 'success' boolean and 'interfaces' list with Django Interface model data
        """
        wireguard_interfaces = []
        wireguard_peers = []
        routes= []
        log.debug(f"Creating wireguard tunnels: {self.tunnels}")
        
        # Get the starting index for new WireGuard interfaces
        current_wg_index = self._get_next_available_wg_index()
        created_tunnels = []
        created_interfaces_data = []  # For Django Interface model data
        
        for tunnel in self.tunnels:
            wg_name = f"wg{current_wg_index}"
            log.debug(f"Creating WireGuard tunnel with name: {wg_name}")
            wireguard_interfaces.append(
                models.InterfaceConfigurationItem(
                    name=wg_name,
                    type=models.InterfaceType.WIREGUARD_TUNNEL,
                    enabled=True,
                    ip_addresses=[f"{tunnel.address}/30"],
                    mtu= tunnel.mtu if tunnel.mtu else 1420,  # default mtu is 1420
                    link=models.WireguardInterfaceLink(
                        private_key=tunnel.private_key,
                        port=tunnel.listen_port,
                        src_addr=ipaddress.IPv4Address(tunnel.source_ip) if tunnel.source_ip else None,
                    ),
                )
            )
            wireguard_peers.append(
                models.WireguardPeerConfigurationItem(
                    wg_if_name=wg_name,
                    public_key=tunnel.peer_public_key,
                    allowed_ips=["0.0.0.0/0"],
                    persistent_keepalive=tunnel.persistent_keepalive,
                    endpoint=ipaddress.IPv4Address(tunnel.peer_endpoint.split(":")[0]) if tunnel.peer_endpoint else None,
                    port=tunnel.peer_endpoint.split(":")[1] if tunnel.peer_endpoint and ":" in tunnel.peer_endpoint else None,
                )
            )
            
            routes.append(
                models.RouteConfigurationItem(
                    destination_network=f"{tunnel.peer_address.split('/')[0]}/32",
                    next_hop_address=tunnel.address.split("/")[0],
                    outgoing_interface=wg_name,
                )
            )
            created_tunnels.append(
                CreatedWireguardTunnel(
                    name=wg_name,  # Use the actual VPP interface name
                    ip_address=tunnel.address,
                    peer_ip_address=tunnel.peer_address,
                    mapped_name=tunnel.interface_name if tunnel.interface_name else None,  # Use the provided interface name if available
                )
            )
            
            # Create Django Interface model data
            ip_address_without_cidr = tunnel.address.split('/')[0]
            subnet_mask = int(tunnel.address.split('/')[1]) if '/' in tunnel.address else 30
            
            interface_data = {
                'name': wg_name,
                'type': 'tunnel',
                'up': True,  # Will be set to True initially, can be verified later
                'hwaddr': None,  # WireGuard interfaces don't have MAC addresses
                'ip_address': ip_address_without_cidr,
                'subnet_mask': subnet_mask,
                'vlan_id': None,
                'mtu': tunnel.mtu if tunnel.mtu else 1420,
                'vpp_used': True,
                'vpp_interface_name': wg_name,
                'is_enabled': True,
                'description': f"WireGuard tunnel to {tunnel.peer_endpoint}",
                'is_management': False,
                'is_wan': True,  # WireGuard tunnels are typically WAN connections
                'is_lan': False,
                'is_primary': False,
                # Additional tunnel-specific metadata
                'tunnel_metadata': {
                    'tunnel_type': 'wireguard',
                    'peer_public_key': tunnel.peer_public_key,
                    'peer_endpoint': tunnel.peer_endpoint,
                    'peer_address': tunnel.peer_address,
                    'listen_port': tunnel.listen_port,
                    'persistent_keepalive': tunnel.persistent_keepalive
                }
            }
            created_interfaces_data.append(interface_data)
            
            current_wg_index += 1
            

        # Apply all configurations at once outside the loop
        if wireguard_interfaces or wireguard_peers or routes:
            reply = self.client.add_configuration(
                *chain(wireguard_interfaces, wireguard_peers, routes)
            )
            log.debug(reply)
            
         
            if reply and reply.processing_error:
                log.error(f"Error adding wireguard tunnels: {reply.processing_error}")
                return {
                    'success': False,
                    'interfaces': [],
                    'error': reply.processing_error
                }
            
            if reply and reply.all_added_items_applied_to_vpp:
                self.wireguard_tunnels = created_tunnels
                log.info(f"Wireguard tunnels created successfully: {[t.name for t in created_tunnels]}")
                return {
                    'success': True,
                    'interfaces': created_interfaces_data,
                    'created_tunnels': [t.name for t in created_tunnels]
                }
            else:
                log.error(f"Failed to create wireguard tunnels")
                return {
                    'success': False,
                    'interfaces': [],
                    'error': 'Failed to apply configuration to VPP'
                }
        else:
            log.warning("No WireGuard tunnels to create")
            return {
                'success': True,
                'interfaces': [],
                'message': 'No tunnels to create'
            }


    def remove_wireguard_tunnels(self) -> bool:
        """
        Remove all WireGuard tunnels.
        Returns:
            bool: True if tunnels were removed successfully, False otherwise.
        """
        if not self.wireguard_tunnels:
            log.warning("No WireGuard tunnels to remove")
            return True
        log.debug(f"Removing WireGuard tunnels: {[t.name for t in self.wireguard_tunnels]}")
        try:
            current_config = self.client.get_configuration()
            configs = []
            
            for tunnel in self.wireguard_tunnels:
                #Find the WireGuard interface configuration
                interface_config = [
                    item.config
                    for item in current_config.items
                    if isinstance(item.config, models.InterfaceConfigurationItem)
                    and item.config.name == tunnel.name
                ]
                if not interface_config:
                    log.error(f"Wireguard interface {tunnel.name} not found")
                    
      
                peer_config = [ item.config  for item in current_config.items if isinstance(item.config, models.WireguardPeerConfigurationItem) and item.config.wg_if_name == tunnel.name]
                if not peer_config:
                    log.error(f"Wireguard peer for interface {tunnel.name} not found")
                    
                
                route_config = [
                    item.config
                    for item in current_config.items
                    if isinstance(item.config, models.RouteConfigurationItem)
                    and item.config.outgoing_interface == tunnel.name
                ]
                if not route_config:
                    log.info(f"Route for Wireguard interface {tunnel.name} not found")
                    
                configs.extend(interface_config + peer_config + route_config)
            if not configs:
                log.warning("No WireGuard configurations found to remove")
                return True
            
            log.debug(f"Configurations to remove: {configs}")
               
            # Remove the WireGuard interface, peer, and route if they exist
            reply = self.client.delete_configuration(*configs)
            log.debug(reply)

            self.wireguard_tunnels = []
            log.info("WireGuard tunnels removed successfully")
            return True
        except Exception as e:
            log.error(f"Exception while removing WireGuard tunnels: {e}")
            return False
        
    def create_gre_tunnels(self) -> List[CreatedWireguardTunnel]:
        """
        Create GRE tunnels based on the WireGuard tunnels.
        Returns:
            List[CreatedWireguardTunnel]: A list of created GRE tunnels.
        """
        if not self.wireguard_tunnels:
            log.error("No WireGuard tunnels created yet")
            return []

        self.gre_tunnels = []
        for i, wg_tunnel in enumerate(self.wireguard_tunnels):
            # Find the corresponding original tunnel config to get GRE fields
            original_tunnel = None
            if i < len(self.tunnels):
                original_tunnel = self.tunnels[i]
            else:
                # Fallback: find by interface name matching
                for tunnel_config in self.tunnels:
                    if tunnel_config.interface_name == wg_tunnel.mapped_name:
                        original_tunnel = tunnel_config
                        break
            log.debug(f"Creating GRE tunnel for WireGuard tunnel: {wg_tunnel.name}, original tunnel: {original_tunnel}")
            gre_name, gre_ip = self.add_wireguard_gre_tunnel(
                wg_tunnel,
                original_tunnel
            )
            self.gre_tunnels.append(
                CreatedWireguardTunnel(name=gre_name, ip_address=gre_ip)
            )
        return self.gre_tunnels
    
    def create_ospf_configuration(self):
        
        # primary_lan = [intf for intf in settings.config.interfaces if intf.is_primary and intf.type == "lan"][0]
        # # expose the primary LAN interface to the FRR container
        # self.expose_lan_interface_to_frr_container(
        #     primary_lan.interface_name,
        #     primary_lan.ip_address,
        #     primary_lan.prefix_len,
            
        # )
        # add bgp loopback so that the FRR container can communicate with the route reflector
        self.add_bgp_loopback(
            settings.config.frr.loopback.split("/")[0]
        )
        if self.ospf_config.client_enabled:
            self.add_ospf_lcp_fix(
                self.ospf_config.lan_ip_address,
                primary_lan.interface_name
            )
            self.add_client_neighbour_info(
                primary_lan.interface_name,
                self.ospf_config.lan_ip_address,
                self.ospf_config.lan_mac_address,
            )
        
        frr = FRR(self.ospf_config, self.wireguard_tunnels )
        frr.apply_all_configs()

        # self.add_routes_for_bidirectional_bgp_communication(
        #     "frr",
        #     settings.config.frr.loopback.split("/")[0],
        #     settings.config.frr.controller_bgp.loopback.split("/")[0],
        #     settings.config.frr.controller_bgp.peer_ip.split("/")[0],
        # )
        
        # self.add_nat_configuration(
        #     lan_iface=settings.config.interfaces["lan_primary"],
        #     wan_iface=settings.config.interfaces["wan_primary"],
        #     wg_ports_to_open=list(
        #         set([tunnel.source_on_host.listen_port for tunnel in self.tunnels])
        #     ),
        #     tcp_ports=[
        #         22,
        #         2206,
        #         2205,
        #     ],  # TODO: Add the ports from the order
        #     udp_ports=[settings.config.route_reflector.port],
        # )
        # self.add_acl_configuration(
        #     ingress_iface=settings.config.interfaces["lan_primary"],
        #     deny_network="8.8.4.4/32",
        # )


    def apply_frr_configuration(self):
        """
        Apply the FRR configuration to the vRouter agent.
        Returns:
            bool: True if the FRR configuration was applied successfully, False otherwise.
        """
        if not self.frr_config:
            log.error("FRR configuration is not set")
            return False
        
        reply = self.client.add_configuration(
            models.FRRConfigurationItem(
                config=self.frr_config,
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(f"Error applying FRR configuration: {reply.processing_error}")
            return False
        return True

        
    def add_lcp_global_configuration(self):
        """
        Add a global LCP configuration to the vRouter agent.
        Args:
            lcp_interface_name (str): The name of the LCP interface.
        Returns:
            bool: True if the LCP configuration was added successfully, False otherwise.
        """
        # Check if the LCP global configuration already exists
        current_config = self.client.get_configuration()
        lcp_config = [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.LCPGlobalsConfigurationItem)
            and item.config.default_namespace == self.lcp_global
        ]
        if lcp_config:
            log.debug("LCP global configuration already exists. Skipping addition.")
            return False
        # Add the LCP global configuration

        reply = self.client.add_configuration(
            models.LCPGlobalsConfigurationItem(
                default_namespace=self.lcp_global,
                lcp_sync=True,
                lcp_auto_subint=True,
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(
                f"Error adding LCP global configuration: {reply.processing_error}"
            )
            return False
        log.info("LCP global configuration added successfully")
        return True

    def delete_lcp_global_configuration(self, namespace: str):
        """
        Delete the global LCP configuration from the vRouter agent.
        Args:
            client: The client object used to communicate with the server.
        Returns:
            bool: True if the LCP configuration was deleted successfully, False otherwise.
        """

        current_config = self.client.get_configuration()
        lcp_config = [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.LCPGlobalsConfigurationItem)
            and item.config.default_namespace == namespace
        ]
        if not lcp_config:
            log.error("LCP global configuration not found")
            return False
        reply = self.client.delete_configuration(lcp_config[0])
        if reply.processing_error:
            log.error(
                f"Error deleting LCP global configuration: {reply.processing_error}"
            )
            return False
        return True

    def add_bgp_loopback(self, ip_addr: str):
        """
        Add a BGP loopback interface to the client.
        Args:
            client: The client object used to communicate with the server.
            ip_addr (str): The IP address of the loopback interface.
        Returns:
            bool: True if the loopback interface was added successfully, False otherwise.
        """

        reply = self.client.add_configuration(
            models.InterfaceConfigurationItem(
                name="loop0",
                type=models.InterfaceType.SOFTWARE_LOOPBACK,
                enabled=True,
                ip_addresses=[ip_addr + "/32"],
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(f"Error adding loopback interface: {reply.processing_error}")
            return False
        return True

    
    def add_vpp_route(self, dest_net, next_hop_addr, outgoing_interface):
        reply = self.client.add_configuration(
            models.RouteConfigurationItem(
                destination_network=dest_net,
                next_hop_address=next_hop_addr,
                outgoing_interface=outgoing_interface,
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(f"Error adding route: {reply.processing_error}")
            return False
        return True

    def add_linux_route(
        self,
        dest_net: str,
        next_hop_addr: str = "",
        outgoing_iface: str = "",
        container: str = "",  # empty container name means default linux namespace
    ):
        reply = self.client.add_configuration(
            models.LinuxRouteConfigurationItem(
                docker_container_name=container,
                destination_network=dest_net,
                next_hop_address=next_hop_addr,
                outgoing_interface=outgoing_iface,
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(f"Error adding route: {reply.processing_error}")
            return False
        return True

 
    def ping(self, destination_ip: str):
        cmd = f"sudo vppctl ping {destination_ip.split('/')[0]} repeat 1"
        res, _ = run_command(cmd.split())
        log.debug(res)
        if "1 sent, 1 received" in res:
            log.info(f"Ping to {destination_ip} successful")
            return True
        log.error(f"Ping to {destination_ip} failed")
        return False

    def check_interface_status(self, interface_name: str) -> bool:
        """
        Check if a VPP interface is up and operational.
        
        Args:
            interface_name (str): Name of the interface to check
            
        Returns:
            bool: True if interface is up, False otherwise
        """
        try:
            cmd = f"sudo vppctl show interface {interface_name}"
            res, _ = run_command(cmd.split())
            log.debug(f"Interface status for {interface_name}: {res}")
            
            # Check if interface is listed and not down
            if interface_name in res and "up" in res.lower() and "down" not in res.lower():
                log.info(f"Interface {interface_name} is up")
                return True
            else:
                log.warning(f"Interface {interface_name} is down or not found")
                return False
                
        except Exception as e:
            log.error(f"Error checking interface {interface_name} status: {e}")
            return False

    def get_interface_configuration(self, interface_name: str) -> Optional[dict]:
        """
        Get detailed configuration information for a VPP interface.
        
        Args:
            interface_name (str): Name of the interface
            
        Returns:
            Optional[dict]: Interface configuration details or None if not found
        """
        try:
            current_config = self.client.get_configuration()
            
            for item in current_config.items:
                if hasattr(item.config, 'name') and item.config.name == interface_name:
                    return {
                        'name': item.config.name,
                        'type': str(item.config.type),
                        'enabled': getattr(item.config, 'enabled', None),
                        'ip_addresses': getattr(item.config, 'ip_addresses', []),
                        'mtu': getattr(item.config, 'mtu', None)
                    }
            
            log.warning(f"Interface {interface_name} not found in configuration")
            return None
            
        except Exception as e:
            log.error(f"Error getting interface {interface_name} configuration: {e}")
            return None

    def test_tunnel_connectivity(self, tunnel_name: str, peer_address: str, retries: int = 3) -> bool:
        """
        Test connectivity through a tunnel by checking interface status and ping.
        
        Args:
            tunnel_name (str): Name of the tunnel interface
            peer_address (str): Peer IP address to ping (without CIDR)
            retries (int): Number of ping attempts
            
        Returns:
            bool: True if tunnel is operational (interface up + ping successful)
        """
        try:
            log.info(f"Testing connectivity for tunnel {tunnel_name} to {peer_address}")
            
            # First check if interface is up
            if not self.check_interface_status(tunnel_name):
                log.error(f"Tunnel {tunnel_name} interface is not up")
                return False
            
            # Test connectivity with retries
            for attempt in range(retries):
                if self.ping(peer_address):
                    log.info(f"Tunnel {tunnel_name} connectivity test successful (attempt {attempt + 1})")
                    return True
                
                if attempt < retries - 1:  # Don't sleep on last attempt
                    log.debug(f"Ping attempt {attempt + 1} failed, retrying...")
                    import time
                    time.sleep(1)
            
            log.error(f"Tunnel {tunnel_name} connectivity test failed after {retries} attempts")
            return False
            
        except Exception as e:
            log.error(f"Error testing tunnel {tunnel_name} connectivity: {e}")
            return False

    def verify_tunnels_operational(self) -> List[dict]:
        """
        Verify all created WireGuard tunnels are operational.
        
        Returns:
            List[dict]: List of tunnel status information
        """
        results = []
        
        if not self.wireguard_tunnels:
            log.warning("No WireGuard tunnels to verify")
            return results
        
        log.info(f"Verifying {len(self.wireguard_tunnels)} WireGuard tunnels")
        
        for tunnel in self.wireguard_tunnels:
            try:
                # Get corresponding tunnel config to find peer address
                tunnel_config = None
                for config in self.tunnels:
                    if config.address.split('/')[0] in tunnel.ip_address:
                        tunnel_config = config
                        break
                
                if not tunnel_config:
                    log.error(f"Could not find config for tunnel {tunnel.name}")
                    results.append({
                        'tunnel_name': tunnel.name,
                        'status': 'error',
                        'interface_up': False,
                        'connectivity': False,
                        'error': 'Configuration not found'
                    })
                    continue
                
                # Check interface status
                interface_up = self.check_interface_status(tunnel.name)
                
                # Test connectivity to peer
                peer_ip = tunnel_config.peer_address.split('/')[0]
                connectivity = False
                if interface_up:
                    connectivity = self.test_tunnel_connectivity(tunnel.name, peer_ip)
                
                status = 'active' if (interface_up and connectivity) else 'inactive'
                
                results.append({
                    'tunnel_name': tunnel.name,
                    'status': status,
                    'interface_up': interface_up,
                    'connectivity': connectivity,
                    'peer_ip': peer_ip,
                    'local_ip': tunnel.ip_address
                })
                
                log.info(f"Tunnel {tunnel.name} verification: {status} (interface: {interface_up}, connectivity: {connectivity})")
                
            except Exception as e:
                log.error(f"Error verifying tunnel {tunnel.name}: {e}")
                results.append({
                    'tunnel_name': tunnel.name,
                    'status': 'error',
                    'interface_up': False,
                    'connectivity': False,
                    'error': str(e)
                })
        
        return results

    def get_wireguard_interfaces(self) -> List[dict]:
        """
        Get list of WireGuard interfaces from VPP configuration.
        
        Returns:
            List[dict]: List of WireGuard interface information
        """
        try:
            # Ensure we're connected before making the call
            if not self.is_connected():
                log.warning("VPP client not connected, attempting to reconnect")
                return []
                
            current_config = self.client.get_configuration()
            wireguard_interfaces = []
            
            for item in current_config.items:
                if (hasattr(item.config, 'type') and 
                    hasattr(models.InterfaceType, 'WIREGUARD_TUNNEL') and
                    item.config.type == models.InterfaceType.WIREGUARD_TUNNEL):
                    
                    iface_info = {
                        'name': item.config.name,
                        'enabled': getattr(item.config, 'enabled', False),
                        'ip_addresses': getattr(item.config, 'ip_addresses', []),
                        'mtu': getattr(item.config, 'mtu', None)
                    }
                    
                    # Extract WireGuard-specific info if available
                    if hasattr(item.config, 'link') and item.config.link:
                        link = item.config.link
                        iface_info.update({
                            'private_key': getattr(link, 'private_key', None),
                            'listen_port': getattr(link, 'port', None),
                            'src_addr': str(getattr(link, 'src_addr', None)) if getattr(link, 'src_addr', None) else None
                        })
                    
                    wireguard_interfaces.append(iface_info)
            
            log.debug(f"Found {len(wireguard_interfaces)} WireGuard interfaces: {[iface['name'] for iface in wireguard_interfaces]}")
            return wireguard_interfaces
            
        except Exception as e:
            log.error(f"Error getting WireGuard interfaces: {e}")
            return []
    
    def get_wireguard_peers(self) -> List[dict]:
        """
        Get list of WireGuard peers from VPP configuration.
        
        Returns:
            List[dict]: List of WireGuard peer information
        """
        try:
            # Ensure we're connected before making the call
            if not self.is_connected():
                log.warning("VPP client not connected, attempting to reconnect")
                return []
                
            current_config = self.client.get_configuration()
            wireguard_peers = []
            
            for item in current_config.items:
                if isinstance(item.config, models.WireguardPeerConfigurationItem):
                    peer_info = {
                        'interface': item.config.wg_if_name,
                        'public_key': item.config.public_key,
                        'allowed_ips': getattr(item.config, 'allowed_ips', []),
                        'endpoint': str(item.config.endpoint) if item.config.endpoint else None,
                        'port': item.config.port,
                        'persistent_keepalive': getattr(item.config, 'persistent_keepalive', None)
                    }
                    wireguard_peers.append(peer_info)
            
            log.debug(f"Found {len(wireguard_peers)} WireGuard peers")
            return wireguard_peers
            
        except Exception as e:
            log.error(f"Error getting WireGuard peers: {e}")
            return []

    def verify_tunnels_operational_enhanced(self, interface_names: Optional[List[str]] = None) -> dict:
        """
        Enhanced version of verify_tunnels_operational that can verify specific interfaces.
        
        Args:
            interface_names: List of interface names to verify. If None, verify all created tunnels.
        
        Returns:
            dict: Mapping of interface names to their verification results
        """
        results = {}
        
        # If specific interface names provided, verify those
        if interface_names:
            log.info(f"Verifying specified WireGuard interfaces: {interface_names}")
            
            for interface_name in interface_names:
                try:
                    # Check interface status
                    interface_up = self.check_interface_status(interface_name)
                    
                    # Get interface configuration to find peer info
                    iface_config = self.get_interface_configuration(interface_name)
                    connectivity_test = False
                    
                    if interface_up and iface_config:
                        # Try to find peer address for connectivity test
                        peers = self.get_wireguard_peers()
                        interface_peers = [p for p in peers if p.get('interface') == interface_name]
                        
                        if interface_peers:
                            # For now, mark connectivity as True if interface is up and has peers
                            # More sophisticated connectivity testing could be added here
                            connectivity_test = True
                    
                    operational = interface_up and connectivity_test
                    
                    results[interface_name] = {
                        'operational': operational,
                        'interface_up': interface_up,
                        'connectivity_test': connectivity_test
                    }
                    
                    log.info(f"Interface {interface_name} verification: operational={operational} "
                           f"(interface_up={interface_up}, connectivity={connectivity_test})")
                    
                except Exception as e:
                    log.error(f"Error verifying interface {interface_name}: {e}")
                    results[interface_name] = {
                        'operational': False,
                        'interface_up': False,
                        'connectivity_test': False,
                        'error': str(e)
                    }
            
            return results
        
        # Fallback to original behavior if no specific interfaces provided
        original_results = self.verify_tunnels_operational()
        
        # Convert to new format
        for result in original_results:
            tunnel_name = result.get('tunnel_name')
            if tunnel_name:
                operational = result.get('status') == 'active'
                results[tunnel_name] = {
                    'operational': operational,
                    'interface_up': result.get('interface_up', False),
                    'connectivity_test': result.get('connectivity', False)
                }
                if 'error' in result:
                    results[tunnel_name]['error'] = result['error']
        
        return results


    def add_ospf_lcp_fix(self, client_vpp_intf_address, vpp_lan_intf):
        # https://lists.fd.io/g/vpp-dev/topic/83103366#19478
        # for 224.0.0.5 and 224.0.0.6 used by OSPF (https://en.wikipedia.org/wiki/Open_Shortest_Path_First)

        # run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", "local", "Forward"])
        # for i in range(4):
        #     run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", os.environ[os.environ["VPP"]+f"_WG{i}_INTERFACE_NAME"], "Accept"])

        # to steer OSPF Hello packets from FRR container to client
        self.add_vpp_route(
            "224.0.0.0/24",  # static due to OSPF hello packets always destined for 224.0.0.5
            client_vpp_intf_address,
            vpp_lan_intf,
        )

        # to get ospf hello packets from client to frr container through vpp
        run_command(
            [
                "sudo",
                "vppctl",
                "ip",
                "mroute",
                "add",
                client_vpp_intf_address,
                "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
                "via",
                vpp_lan_intf,
                "Accept",
            ]
        )
        run_command(
            [
                "sudo",
                "vppctl",
                "ip",
                "mroute",
                "add",
                client_vpp_intf_address,
                "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
                "via",
                "local",  # static, set to be locally processed (processed by LCP)
                "Forward",
            ]
        )

    # Note: this does not include RR-node wireguard tunnel (for that see makefile) and RR setup => only path from linux WG
    # on node to BGP in container on node
    def add_routes_for_bidirectional_bgp_communication(
        self,
        local_bgp_loop_ip: str,
        rr_bgp_loop_ip: str,
        rr_wg_ip: str,
        frr_container_name: str = "frr"
    ):
        # Some path explanation:
        # RR-to-node path: node-linux-wg0 -(adding here route1+iptables forwarding)-> FRR container default interface ->
        #   FRR container routing to local loopback -(ping echo going back, need to add route2)-> FRR container default
        #   interface -(routing path was added as part of wg configuration)-> node-linux-wg0
        # Node-to-RR path: FRR instance in FRR container -(need to add route3)-> FRR container default interface
        #   -> node linux host route(actually controlled by WG allowed ip configuration) -> wg tunnel to RR
        host_ip_from_docker_container = "172.17.0.1"

        # adding route1
        frr_container_ip_address, _ = run_command(
            [
                "docker",
                "inspect",
                "-f",
                "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}'",
                frr_container_name,
            ]
        )
        frr_container_ip_address = frr_container_ip_address.replace("\n", "").replace(
            "'", ""
        )
        log.debug(f"FRR container IP address: {frr_container_ip_address}")

        self.add_linux_route(
            local_bgp_loop_ip,
            frr_container_ip_address,
        )
        run_command(
            [
                "sudo",
                "iptables",
                "-A",
                "FORWARD",
                "-i",
                "wg0",  # static for default wireguard interface between controller and node 
                "-o",
                "docker0",  # static for default docker installation
                "-j",
                "ACCEPT",
            ]
        )
        # adding route2
        self.add_linux_route(
            rr_wg_ip,
            host_ip_from_docker_container,
            "",
            frr_container_name,
        )
        # adding route3
        self.add_linux_route(
            rr_bgp_loop_ip,
            host_ip_from_docker_container,
            "",
            frr_container_name,
        )

    def add_client_neighbour_info(
        self, vpp_lan_intf: str, client_vpp_intf_ip: str, client_vpp_intf_mac: str
    ):

        # using MAC address of client interface (or at least what TREX traffic destination MAC address that VPP must
        # steer back to Trex through client interface)
        cmd = f"sudo vppctl ip neighbor {vpp_lan_intf} {client_vpp_intf_ip} {client_vpp_intf_mac}"
        _, ret = run_command(cmd.split())
        if ret == 0:
            log.info(f"Added neighbour info for {client_vpp_intf_ip}")
        else:

            log.error(f"Error adding neighbour info for {client_vpp_intf_ip}")
            return False
        return True

    def add_wireguard_gre_tunnel(
        self,
        tunnel: CreatedWireguardTunnel,
        original_tunnel_config: Optional[WireGuardConfig] = None,
    ):
        """
        Add a GRE tunnel for a WireGuard tunnel using predefined GRE configuration or fallback to automatic generation.
        
        Args:
            tunnel: WireGuard tunnel configuration
            original_tunnel_config: Original tunnel configuration containing GRE fields
            
        Returns:
            Tuple of (gre_interface_name, gre_interface_ip_with_mask)
        """
        try:
            # Check if we have predefined GRE configuration
            if (original_tunnel_config and 
                original_tunnel_config.gre_name and 
                original_tunnel_config.gre_local_tunnel_ip and 
                original_tunnel_config.gre_remote_tunnel_ip):
                
                # Use predefined GRE configuration from tunnel data
                gre_interface_name = original_tunnel_config.gre_name
                gre_interface_ip_address = original_tunnel_config.gre_local_tunnel_ip
                remote_gre_interface_ip_address = original_tunnel_config.gre_remote_tunnel_ip
                
                # Determine the subnet mask - default to /30 for GRE point-to-point links
                if "/" in gre_interface_ip_address:
                    local_gre_ip_with_mask = gre_interface_ip_address
                    gre_interface_ip_address = gre_interface_ip_address.split("/")[0]
                    gre_interface_ip_address_mask = local_gre_ip_with_mask.split("/")[1]
                else:
                    # Default to /30 for point-to-point GRE tunnels
                    gre_interface_ip_address_mask = "30"
                    local_gre_ip_with_mask = f"{gre_interface_ip_address}/{gre_interface_ip_address_mask}"
                
                log.info(
                    f"Using predefined GRE configuration - Interface: {gre_interface_name}, "
                    f"Local: {local_gre_ip_with_mask}, Remote: {remote_gre_interface_ip_address}, "
                    f"WG Tunnel: {tunnel.name} ({tunnel.ip_address})"
                )
            else:
                # Fallback to automatic generation using GRE manager
                log.info(f"No predefined GRE configuration found for {tunnel.name}, using automatic generation")
                
                # Extract index from tunnel name (e.g., "wg0" -> "0")
                if tunnel.name.startswith("wg"):
                    index = tunnel.name[2:]  # Remove "wg" prefix to get the numeric index
                else:
                    # Fallback to old method for backward compatibility
                    index = tunnel.name.split("_")[-1]
                
                gre_interface_name = f"gre{index}"
                
                # Use the new GRE manager to allocate addresses safely
                gre_manager = get_gre_manager()
                local_gre_ip_with_mask, remote_gre_interface_ip_address = convert_wg_to_gre_ip_addr_safe(tunnel, gre_manager)
                
                # Extract components for processing
                gre_interface_ip_address = local_gre_ip_with_mask.split("/")[0]
                gre_interface_ip_address_mask = local_gre_ip_with_mask.split("/")[1]
                
                log.info(
                    f"GRE tunnel allocation (auto-generated) - Interface: {gre_interface_name}, "
                    f"Local: {local_gre_ip_with_mask}, Remote: {remote_gre_interface_ip_address}, "
                    f"WG Tunnel: {tunnel.name} ({tunnel.ip_address})"
                )

            def add_frr_agent_ip_translation_from_gre_to_wg(
                gre_interface_ip_address: str,
                interface_ip: str,
                remote_gre_interface_ip_address: str,
                remote_tunnel_end_ip: str,
                frr_container_name: str = "frr",
                replace_file_path: str = "/rr-ip-replace.txt",
            ):
                # TODO when proper python client API will be created for this then replace this with python client code
                _, code1 = run_command(
                    [
                        "docker",
                        "exec",
                        "-t",
                        frr_container_name,
                        "sh",
                        "-c",
                        "echo '"
                        + gre_interface_ip_address
                        + "=>"
                        + interface_ip
                        + "' >> "
                        + replace_file_path,
                    ]
                )
                _, code2 = run_command(
                    [
                        "docker",
                        "exec",
                        "-t",
                        frr_container_name,
                        "sh",
                        "-c",
                        "echo '"
                        + remote_gre_interface_ip_address
                        + "=>"
                        + remote_tunnel_end_ip
                        + "' >> "
                        + replace_file_path,
                    ]
                )
                return code1 == 0 and code2 == 0

            def add_lcp_interface(
                vpp_iface: str,
                host_iface: str,
                host_netns: str,
                is_tun: bool = False,
            ):
                iface_type = (
                    models.LCPHostInterfaceTypeEnum.TUN
                    if is_tun
                    else models.LCPHostInterfaceTypeEnum.TAP
                )
                reply = self.client.add_configuration(
                    models.LCPPairConfigurationItem(
                        interface=vpp_iface,
                        mirror_interface_host_name=host_iface,
                        mirror_interface_type=iface_type,
                        host_namespace=host_netns,
                    )
                )
                log.debug(reply)
                if reply.processing_error:
                    log.error(f"Error adding LCP interface: {reply.processing_error}")
                    return False
                return True

            def add_gre_tunnel_to_wg_and_lcp_it_to_frr_container(
                gre_interface_name: str,
                gre_interface_ip_address: str,
                gre_interface_ip_address_mask: str,
                remote_gre_interface_ip_address: str,
                interface_name: str,
                remote_tunnel_end_ip: str,
                frr_container_name: str = "frr",
            ):

                gre_iface = models.InterfaceConfigurationItem(
                    name=gre_interface_name,
                    type=models.InterfaceType.GRE_TUNNEL,
                    enabled=False,
                    ip_addresses=[],
                    link=models.GREInterfaceLink(
                        type=models.GRELinkType.L3,
                        src_addr=ipaddress.IPv4Address(gre_interface_ip_address),
                        dst_addr=ipaddress.IPv4Address(remote_gre_interface_ip_address),
                    ),
                )

                reply = self.client.add_configuration(gre_iface)
                if reply and reply.processing_error:
                    log.error(f"Error adding GRE interface: {reply.processing_error}")
                    return False

                self.add_vpp_route(
                    remote_gre_interface_ip_address
                    + "/32",  # route to one ip address => mask /32 is static value
                    remote_tunnel_end_ip,
                    interface_name,
                )

                add_lcp_interface(
                    gre_interface_name,
                    interface_name,
                    frr_container_name,
                    True,
                )

                reply = self.client.update_configuration(
                    gre_iface,
                    models.InterfaceConfigurationItem(
                        name=gre_interface_name,
                        type=models.InterfaceType.GRE_TUNNEL,
                        enabled=True,
                        ip_addresses=[
                            gre_interface_ip_address + "/" + gre_interface_ip_address_mask
                        ],
                        link=models.GREInterfaceLink(
                            type=models.GRELinkType.L3,
                            src_addr=ipaddress.IPv4Address(gre_interface_ip_address),
                            dst_addr=ipaddress.IPv4Address(remote_gre_interface_ip_address),
                        ),
                    ),
                )
                log.debug(reply)
                if reply and reply.processing_error:
                    log.error(f"Error enabling GRE interface: {reply.processing_error}")
                    return False
                return True

            if add_gre_tunnel_to_wg_and_lcp_it_to_frr_container(
                gre_interface_name,
                gre_interface_ip_address,
                gre_interface_ip_address_mask,
                remote_gre_interface_ip_address,
                tunnel.name,
                self._safe_extract_ip(tunnel.peer_ip_address) if hasattr(tunnel, 'peer_ip_address') and tunnel.peer_ip_address else remote_gre_interface_ip_address,
            ):
                log.info(f"Added GRE tunnel to VPP for {gre_interface_name}")

            run_command(
                [
                    "sudo",
                    "vppctl",
                    "ip",
                    "mroute",
                    "add",
                    remote_gre_interface_ip_address,
                    "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
                    "via",
                    gre_interface_name,
                    "Accept",
                ]
            )
            run_command(
                [
                    "sudo",
                    "vppctl",
                    "ip",
                    "mroute",
                    "add",
                    remote_gre_interface_ip_address,
                    "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
                    "via",
                    "local",  # static forwarding to local processing (it will be processed by LCP)
                    "Forward",
                ]
            )

            if add_frr_agent_ip_translation_from_gre_to_wg(
                gre_interface_ip_address,
                self._safe_extract_ip(tunnel.ip_address),
                remote_gre_interface_ip_address,
                self._safe_extract_ip(tunnel.peer_ip_address) if hasattr(tunnel, 'peer_ip_address') and tunnel.peer_ip_address else remote_gre_interface_ip_address,
            ):
                log.info(
                    f"Added FRR agent IP translation for GRE interface {gre_interface_name}"
                )
            
            return (
                gre_interface_name,
                local_gre_ip_with_mask,
            )
            
        except Exception as e:
            log.error(f"Failed to create GRE tunnel for {tunnel.name}: {e}")
            # Attempt cleanup on failure only if we used automatic generation
            if not (original_tunnel_config and 
                    original_tunnel_config.gre_name and 
                    original_tunnel_config.gre_local_tunnel_ip and 
                    original_tunnel_config.gre_remote_tunnel_ip):
                try:
                    gre_manager = get_gre_manager()
                    gre_manager.deallocate_gre_addresses(tunnel.name)
                except Exception as cleanup_error:
                    log.warning(f"Failed to cleanup GRE allocation for {tunnel.name}: {cleanup_error}")
            raise

    def expose_lan_interface_to_frr_container(
        self,
        interface: Interface,
        frr_container_name: str = "frr",
    ):
        lcp = f"sudo vppctl lcp create {interface.interface_name} host-if {interface.interface_name} netns {frr_container_name} tun"
        _, ret1 = run_command(lcp.split())
        if ret1 == 0:
            log.debug(f"Created LCP interface: {interface.interface_name}")
        else:
            log.error(f"Error creating LCP interface: {interface.interface_name}")
            return False

        ip_addr_del = f"sudo vppctl set interface ip address del {interface.interface_name} {str(interface.ip_address)}/{interface.prefix_len}"
        _, ret2 = run_command(ip_addr_del.split())

        if ret2 == 0:
            log.debug(f"Deleted IP address from interface: {interface.interface_name}")
        else:
            log.error(f"Error deleting IP address from interface: {interface.interface_name}")
            return False

        set_ip = f"sudo vppctl set interface ip address {interface.interface_name} {str(interface.ip_address)}/{interface.prefix_len}"
        _, ret3 = run_command(set_ip.split())

        if ret3 == 0:
            log.debug(f"Set IP address on interface: {interface.interface_name}")
        else:
            log.error(f"Error setting IP address on interface: {interface.interface_name}")
            return False

    def add_nat_configuration(
        self,
        lan_iface: Interface,
        wan_iface: Interface,
        wg_ports_to_open: List[int],
        tcp_ports: List[int] = [],
        udp_ports: List[int] = [],
    ):
        lan_nat_iface = models.Nat44InterfaceConfigurationItem(
            name=lan_iface.interface_name,
            nat_inside=True,
            nat_outside=False,
            output_feature=False,
        )
        wan_nat_iface = models.Nat44InterfaceConfigurationItem(
            name=wan_iface.interface_name,
            nat_inside=False,
            nat_outside=True,
            output_feature=False,
        )
        nat_address_pool = models.Nat44AddressPoolConfigurationItem(
            name="nat-pool",
            first_ip=ipaddress.IPv4Address(wan_iface.ip_address),
            last_ip=ipaddress.IPv4Address(wan_iface.ip_address),
        )

        wg_port_mappings = [
            models.IdentityMapping(
                interface=wan_iface.interface_name,
                protocol=models.ProtocolInNAT.UDP,
                port=p,
            )
            for p in wg_ports_to_open
        ]
        tcp_port_mappings = [
            models.IdentityMapping(
                interface=wan_iface.interface_name,
                protocol=models.ProtocolInNAT.TCP,
                port=p,
            )
            for p in tcp_ports
        ]
        udp_port_mappings = [
            models.IdentityMapping(
                interface=wan_iface.interface_name,
                protocol=models.ProtocolInNAT.UDP,
                port=p,
            )
            for p in udp_ports
        ]

        nat_mappings = models.DNat44ConfigurationItem(
            label="nat-mappings",
            static_mappings=[],
            identity_mappings=list(
                chain(wg_port_mappings, tcp_port_mappings, udp_port_mappings)
            ),
        )
        reply = self.client.add_configuration(
            lan_nat_iface,
            wan_nat_iface,
            nat_address_pool,
            nat_mappings,
        )
        log.debug(reply)
        if reply and reply.processing_error:
            log.error(f"Error adding NAT configuration: {reply.processing_error}")
            return False
        return True

    def add_acl_configuration(
        self,
        ingress_iface: Interface,
        deny_network: str,
    ):
        reply = self.client.add_configuration(
            models.ACLConfigurationItem(
                name="lab-acl-rules",
                ingress=[ingress_iface.interface_name],
                rules=[
                    models.ACLRuleConfigurationItem(
                        action=models.ACLAction.DENY,
                        refinement=models.IPSpecification(
                            addresses=models.IPAddresses(
                                destination_network=ipaddress.IPv4Network(deny_network)
                            )
                        ),
                    ),
                    models.ACLRuleConfigurationItem(
                        action=models.ACLAction.PERMIT,
                    ),
                ],
            )
        )
        log.debug(reply)
        if reply and reply.processing_error:
            log.error(f"Error adding ACL configuration: {reply.processing_error}")
            return False
        return True

    def delete_nat_configuration(
        self,
        lan_iface: Interface,
        wan_iface: Interface,
        wg_ports_to_open: List[int],
        tcp_ports: List[int] = [],
        udp_ports: List[int] = [],
    ):
        current_config = self.client.get_configuration()
        nat_config = [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.Nat44InterfaceConfigurationItem)
            and item.config.name in [lan_iface.interface_name, wan_iface.interface_name]
        ]
        port_mappings = [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.IdentityMapping)
            and item.config.port in chain(wg_ports_to_open, tcp_ports, udp_ports)
        ]

        if not nat_config:
            log.error("NAT configuration not found")
            return False
        reply = self.client.delete_configuration(*nat_config, *port_mappings)
        if reply.processing_error:
            log.error(f"Error deleting NAT configuration: {reply.processing_error}")
            return False
        log.info("NAT configuration deleted successfully")
        return True

    def delete_acl_configuration(self, lan_iface):
        current_config = self.client.get_configuration()
        acl_config = [
            item.config
            for item in current_config.items
            if isinstance(item.config, models.ACLConfigurationItem)
            and item.config.ingress == [lan_iface.interface_name]
        ]
        if not acl_config:
            log.error("ACL configuration not found")
            return False
        reply = self.client.delete_configuration(*acl_config)
        if reply.processing_error:
            log.error(f"Error deleting ACL configuration: {reply.processing_error}")
            return False
        log.info("ACL configuration deleted successfully")
        return True

    def remove_gre_tunnels(self) -> bool:
        """
        Remove GRE tunnels and deallocate their addresses.
        
        Returns:
            True if successful, False otherwise
        """
        try:
            gre_manager = get_gre_manager()
            success = True
            
            # Clean up GRE allocations for all WireGuard tunnels
            if self.wireguard_tunnels:
                for tunnel in self.wireguard_tunnels:
                    if not gre_manager.deallocate_gre_addresses(tunnel.name):
                        log.warning(f"Failed to deallocate GRE addresses for {tunnel.name}")
                        success = False
            
            # Clear the GRE tunnels list
            self.gre_tunnels = []
            
            if success:
                log.info("GRE tunnels removed and addresses deallocated successfully")
            else:
                log.warning("GRE tunnel removal completed with some warnings")
                
            return success
            
        except Exception as e:
            log.error(f"Exception while removing GRE tunnels: {e}")
            return False
    
    def _safe_extract_ip(self, ip_with_mask: str) -> str:
        """
        Safely extract IP address from IP/mask format.
        
        Args:
            ip_with_mask: IP address potentially with CIDR notation
            
        Returns:
            IP address without CIDR notation
        """
        if not ip_with_mask:
            return "0.0.0.0"  # Fallback
        
        if "/" in ip_with_mask:
            return ip_with_mask.split("/")[0]
        else:
            return ip_with_mask



import ipaddress
import time

from vpp_vrouter.client import ExtendedVPPAPIClient
from vpp_vrouter.common import models

from vrouter_agent import utils
from vrouter_agent.utils import run_command

logger = utils.setup_logger(__name__, save_dir="/var/log")


def configure_dpdk_interfaces(
    client,
    wan_intf_name: str,
    lan_intf_name: str,
    wan_intf_net: list[str] = [],
    lan_intf_net: list[str] = [],
):
    wan_intf_config = models.InterfaceConfigurationItem(
        name=wan_intf_name,
        type=models.InterfaceType.DPDK,
        enabled=True,
        ip_addresses=[wan_intf_net],
        mtu=8500,
    )
    lan_intf_config = models.InterfaceConfigurationItem(
        name=lan_intf_name,
        type=models.InterfaceType.DPDK,
        enabled=True,
        ip_addresses=[lan_intf_net],
        mtu=8500,
    )
    # Configure DPDK interfaces with 10.0.x.1/24 IP addresses
    for item in client.get_configuration().items:
        # Make sure the interfaces are configured
        if (
            isinstance(item.config, models.InterfaceConfigurationItem)
            and item.config.name == wan_intf_name
            and not item.config.enabled
        ):
            logger.info("Enabling WAN interface %s" % wan_intf_name)
            reply = client.update_configuration(
                item.config,
                wan_intf_config,
            )
            if reply.all_updated_items_applied_to_vpp:
                logger.info(f"WAN interface {item.config.name} configured")

        if (
            isinstance(item.config, models.InterfaceConfigurationItem)
            and item.config.name == lan_intf_name
            and not item.config.enabled
        ):
            logger.info("Enabling LAN interface %s" % lan_intf_name)
            reply = client.update_configuration(
                item.config,
                lan_intf_config,
            )
            if reply.all_updated_items_applied_to_vpp:
                logger.info(f"LAN interface {item.config.name} configured")


def configure_wg_interfaces(
    client,
    wg_name: str,
    wg_net: list[str],
    mtu: int,
    src_addr: str,
    src_port: int,
    dst_port: int,
    private_key: str,
    dst_addr: str,
    peer_public_key: str,
    peer_endpoint: str,
    peer_allowed_ips: list[str],
):
    logger.info("Configuring Wireguard")
    wireguard_interface_config = models.InterfaceConfigurationItem(
        name=wg_name,
        enabled=True,
        ip_addresses=wg_net,
        type=models.InterfaceType.WIREGUARD_TUNNEL,
        mtu=mtu,  # leaving some room for encryption encapsulation so that packet don't break into segments
        link=models.WireguardInterfaceLink(
            private_key=private_key,  # encoded in base64
            port=src_port,
            src_addr=ipaddress.IPv4Address(src_addr),
        ),
    )
    wireguard_peer_config = models.WireguardPeerConfigurationItem(
        public_key=peer_public_key,  # encoded in base64
        port=dst_port,
        endpoint=ipaddress.IPv4Address(dst_addr),
        allowed_ips=peer_allowed_ips,
        wg_if_name=wireguard_interface_config.name,
    )
    wireguard_router_config = models.RouteConfigurationItem(
        destination_network=peer_endpoint,
        outgoing_interface=wireguard_interface_config.name,
    )
    reply = client.add_configuration(
        wireguard_interface_config, wireguard_peer_config, wireguard_router_config
    )
    if reply.vpp_apply_success:
        logger.info("Wireguard configured. Now wait 30s for the tunnel to come up")
        time.sleep(30)
        logger.info("Testing connectivity")
        output = run_command(["sudo", "vppctl", "ping", peer_endpoint.split("/")[0]])
        logger.info(output)
    else:
        logger.info("Wireguard configuration failed")


def configure_nat(client, wan_intf_name, lan_intf_name):
    logger.info("Configuring NAT")
    # NAT INSIDE
    nat44_in_config = models.Nat44InterfaceConfigurationItem(
        name=lan_intf_name, nat_inside=True
    )
    # NAT OUTSIDE
    nat44_out_config = models.Nat44InterfaceConfigurationItem(
        name=wan_intf_name, nat_outside=True
    )
    # IDENTITY NAT
    indentity_nat_config = models.DNat44ConfigurationItem(
        label="identity_nat",
        identity_mappings=[models.IdentityMapping(interface=wan_intf_name, port=50000)],
    )

    # ADDRESS POOL
    # address_pool_config = models.Nat44AddressPoolConfigurationItem(
    #     first_ip="10.0.1.1",
    #     last_ip="10.0.2.1",
    # )

    # Create temp cli command for NAT
    run_command(["sudo", "vppctl", "nat44" "add", "10.0.2.1"])
    run_command(["sudo", "vppctl", "nat44", "add", "10.0.1.1"])

    reply = client.add_configuration(
        nat44_in_config, nat44_out_config, indentity_nat_config
    )
    logger.info(reply)

    logger.info("NAT configured. Now wait 30s for the tunnel to come up")
    time.sleep(10)

    logger.info("Testing connectivity")
    output = run_command(["sudo", "vppctl", "ping", "172.32.0.10"])
    logger.info(output)


def configure_lcp(client, wan_intf_name, lan_intf_name):
    # Configuring LCP
    logger.info("Configuring LCP")

    lcp_globals_config = models.LCPGlobalsConfigurationItem(
        default_namespace="dataplane",
        lcp_sync=True,  # Enables copying of changes made in VPP into their Linux counterpart
        # sub-interface creation in VPP automatically creates a Linux Interface Pair(LIP) and its companion Linux
        # network interface
        lcp_auto_subint=True,
    )
    lcp_pair_wan_config = models.LCPPairConfigurationItem(
        interface=wan_intf_name,
        mirror_interface_host_name=f"host-wan-vpp",
        mirror_interface_type=models.LCPHostInterfaceTypeEnum.TUN,
        host_namespace="dataplane",  # current/default linux namespace
    )
    lcp_pair_lan_config = models.LCPPairConfigurationItem(
        interface=lan_intf_name,
        mirror_interface_host_name=f"host-lan-vpp",
        mirror_interface_type=models.LCPHostInterfaceTypeEnum.TUN,
        host_namespace="dataplane",  # current/default linux namespace
    )

    reply = client.add_configuration(
        lcp_globals_config, lcp_pair_wan_config, lcp_pair_lan_config
    )
    print(reply)


def configure_frr(client):
    # Testing FRR

    logger.info("Configuring FRR")

    frr_init_config = [
        config_detail.config
        for config_detail in client.get_configuration().items
        if isinstance(config_detail.config, models.FRRConfigurationItem)
    ][0]
    print(frr_init_config)

    # appending some bgp configuration to initial FRR config
    # (FRR config is updated, but that means just means delete of old config and add of new one. In case of FRR,
    # the delete does nothing because FRR need to have always some configuration. That means that only addition of
    # FRR configuration matters.)
    append_to_frr_config = (
        """ 
hostname ospfd
password zebra
log stdout
!
router ospf
 network 172.16.0.2/32 area 0
 network 10.0.10.2/24 area 0
 network 10.0.20.2/24 area 0
! 
    """.strip(
            "\n "
        )
        + "\n"
    )

    reply = client.update_configuration(
        frr_init_config,
        models.FRRConfigurationItem(
            config=frr_init_config.config + append_to_frr_config
        ),
    )

    print(reply)


def run():
    with ExtendedVPPAPIClient() as client:
        # DPDK interfaces
        wan_intf_name = "GigabitEthernet0/9/0"
        lan_intf_name = "GigabitEthernet0/8/0"
        wan_intf_net = "10.0.1.2/24"
        lan_intf_net = "10.0.2.2/24"

        # WG Interface
        wg_name = "wg0"
        wg_net = ["172.32.0.9/30"]
        mtu = 1420
        src_port = 50000
        dst_port = 50000
        private_key = "YIBI+xgvhGiTwwblC9DA4Xy3jAyMP2eVVtvst6Z7lmA="
        src_addr = "10.0.1.2"

        # WG Peers
        dst_addr = "10.0.1.1"
        peer_allowed_ips = ["0.0.0.0/0"]
        peer_public_key = "slDc60lMEXVJs2E4MILoSqdqs7DYkrFkGoon8dkqtzI="
        peer_endpoint = "172.32.0.10/30"

        configure_dpdk_interfaces(
            client, wan_intf_name, lan_intf_name, wan_intf_net, lan_intf_net
        )
        configure_wg_interfaces(
            client,
            wg_name=wg_name,
            wg_net=wg_net,
            mtu=mtu,
            src_port=src_port,
            dst_port=dst_port,
            private_key=private_key,
            src_addr=src_addr,
            dst_addr=dst_addr,
            peer_allowed_ips=peer_allowed_ips,
            peer_public_key=peer_public_key,
            peer_endpoint=peer_endpoint,
        )
        # configure_nat(client, wan_intf_name, lan_intf_name)
        configure_lcp(client, wan_intf_name, lan_intf_name)
        configure_frr(client)



"""
Tunnel, Topology, and Network models for FastAPI SQLModel.
"""

from typing import Optional, List, Dict, Any
from sqlmodel import SQLModel, Field, Relationship
from enum import Enum
from .base import BaseModel


class TopologyType(str, Enum):
    HUB_SPOKE = "hub_spoke"
    POINT_TO_POINT = "point_to_point"
    FULL_MESH = "full_mesh"
    PARTIAL_MESH = "partial_mesh"


class TunnelType(str, Enum):
    WIREGUARD = "wireguard"
    IPSEC = "ipsec"
    GRE = "gre"
    VXLAN = "vxlan"
    OPENVPN = "openvpn"


class TunnelState(str, Enum):
    PLANNED = "planned"
    CONFIGURED = "configured"
    ACTIVE = "active"
    INACTIVE = "inactive"
    ERROR = "error"


class Network(BaseModel, table=True):
    """Network model equivalent to Django Network model"""

    __tablename__ = "networks"

    name: str = Field(max_length=100)
    cidr: str = Field(max_length=18, description="Network CIDR (e.g., 192.168.1.0/24)")
    description: Optional[str]

    # Relationships
    tunnels: List["Tunnel"] = Relationship(back_populates="tunnel_network")


class Topology(BaseModel, table=True):
    """Topology model equivalent to Django Topology model"""

    __tablename__ = "topologies"

    name: str = Field(max_length=100)
    topology_type: TopologyType
    description: Optional[str]
    hub_node_id: Optional[int] = Field(foreign_key="devices.id")
    is_active: bool = Field(default=True)

    # Relationships
    hub_node: Optional["Device"] = Relationship(
        sa_relationship_kwargs={
            "foreign_keys": "Topology.hub_node_id",
            "post_update": True,
        }
    )
    tunnels: List["Tunnel"] = Relationship(back_populates="topology")
    orders: List["Order"] = Relationship(back_populates="topology")


class Tunnel(BaseModel, table=True):
    """Tunnel model equivalent to Django Tunnel model"""

    __tablename__ = "tunnels"

    name: Optional[str] = Field(max_length=100)
    tunnel_type: TunnelType

    # Relationships
    topology_id: Optional[int] = Field(foreign_key="topologies.id")
    topology: Optional[Topology] = Relationship(back_populates="tunnels")

    # Node relationships
    local_node_id: Optional[int] = Field(foreign_key="devices.id")
    remote_node_id: Optional[int] = Field(foreign_key="devices.id")
    local_node: Optional["Device"] = Relationship(
        back_populates="local_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_node_id"},
    )
    remote_node: Optional["Device"] = Relationship(
        back_populates="remote_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_node_id"},
    )

    # Interface relationships
    local_interface_id: Optional[int] = Field(foreign_key="interfaces.id")
    remote_interface_id: Optional[int] = Field(foreign_key="interfaces.id")
    local_wan_interface_id: Optional[int] = Field(foreign_key="interfaces.id")
    remote_wan_interface_id: Optional[int] = Field(foreign_key="interfaces.id")

    local_interface: Optional["Interface"] = Relationship(
        back_populates="local_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_interface_id"},
    )
    remote_interface: Optional["Interface"] = Relationship(
        back_populates="remote_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_interface_id"},
    )
    local_wan_interface: Optional["Interface"] = Relationship(
        back_populates="local_wan_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.local_wan_interface_id"},
    )
    remote_wan_interface: Optional["Interface"] = Relationship(
        back_populates="remote_wan_tunnels",
        sa_relationship_kwargs={"foreign_keys": "Tunnel.remote_wan_interface_id"},
    )

    # Network relationship
    tunnel_network_id: Optional[int] = Field(foreign_key="networks.id")
    tunnel_network: Optional[Network] = Relationship(back_populates="tunnels")

    # IP configuration
    local_tunnel_ip: Optional[str] = Field(description="Local tunnel IP address")
    remote_tunnel_ip: Optional[str] = Field(description="Remote tunnel IP address")

    # WireGuard specific settings
    local_public_key: Optional[str]
    local_private_key: Optional[str]
    remote_public_key: Optional[str]
    remote_private_key: Optional[str]  # Temporarily stored for blockchain publishing
    listen_port: Optional[int] = Field(ge=1, le=65535)

    # General tunnel settings
    mtu: int = Field(default=1420)
    keepalive_interval: int = Field(default=25)
    state: TunnelState = Field(default=TunnelState.PLANNED)
    is_enabled: bool = Field(default=True)


# Import here to avoid circular imports
from .devices import Device, Interface
from .orders import Order

# Rebuild models to resolve forward references
Network.model_rebuild()
Topology.model_rebuild()
Tunnel.model_rebuild()



import os
from pathlib import Path
from loguru import logger
from sqlmodel import SQLModel, create_engine, Session

# from vrouter_agent.utils.config import settings


# Use configurable path for database to support different environments
def get_db_dir():
    """Get database directory based on environment variables or defaults."""
    # Check if we're in a test environment
    if os.environ.get("PYTEST_CURRENT_TEST") or os.environ.get("VROUTER_AGENT_ENV") == "test":
        # Use current directory for tests
        db_dir = Path(os.environ.get("VROUTER_AGENT_DB_DIR", "./test-data"))
    else:
        # Use system directory for production
        db_dir = Path(os.environ.get("VROUTER_AGENT_DB_DIR", "/var/db/vrouter-agent/data"))
    
    db_dir.mkdir(parents=True, exist_ok=True)  # Create data directory if it doesn't exist
    return db_dir

# Initialize database configuration lazily
_db_dir = None
_db_file = None
_database_url = None
_engine = None

def get_database_url():
    """Get database URL, initializing if necessary."""
    global _database_url, _db_dir, _db_file
    if _database_url is None:
        _db_dir = get_db_dir()
        _db_file = _db_dir / "data.db"
        _database_url = f"sqlite:///{_db_file}"
    return _database_url

def get_engine():
    """Get database engine, creating if necessary."""
    global _engine
    if _engine is None:
        database_url = get_database_url()
        _engine = create_engine(
            database_url,
            pool_size=10,  # Increase the pool size
            max_overflow=20,  # Increase the overflow size
            pool_timeout=60,  # Increase the connection timeout
            echo=False,  # Set to True for SQL debugging
        )
    return _engine

# Backward compatibility - direct function assignments instead of properties
DB_DIR = get_db_dir

def DB_FILE():
    global _db_file
    if _db_file is None:
        get_database_url()  # This will initialize _db_file
    return _db_file

DATABASE_URL = get_database_url

# Create engine attribute that evaluates to the actual engine object
# This resolves the issue where @property was creating a property object
# Now engine will be the actual SQLAlchemy engine instance
engine = get_engine()

# SQLite database connection
def init_db():
    """
    Initialize database tables.
    This function is safe to call multiple times - it only creates tables
    that don't already exist and preserves all existing data.
    """
    try:
        # Ensure database directory exists
        db_dir = get_db_dir()
        db_dir.mkdir(parents=True, exist_ok=True)
        
        # Get engine and create tables
        engine_instance = get_engine()
        
        # Create tables only if they don't exist (preserves existing data)
        SQLModel.metadata.create_all(engine_instance)
        
        db_file = get_database_url().replace("sqlite:///", "")
        logger.info(f"Database initialized at: {db_file}")
        logger.debug(f"Database URL: {get_database_url()}")
        
        # Log database size if it exists
        db_file_path = Path(db_file)
        if db_file_path.exists():
            size_mb = db_file_path.stat().st_size / (1024 * 1024)
            logger.debug(f"Database size: {size_mb:.2f} MB")
            
    except Exception as e:
        logger.error(f"Failed to initialize database: {e}")
        raise


def get_session():
    """
    Get database session with proper error handling.
    """
    try:
        with Session(get_engine()) as session:
            yield session
    except Exception as e:
        logger.error(f"Database session error: {e}")
        raise


def get_database_info():
    """
    Get information about the current database.
    """
    try:
        db_file_path = Path(get_database_url().replace("sqlite:///", ""))
        db_dir_path = get_db_dir()
        
        info = {
            "database_url": get_database_url(),
            "database_file": str(db_file_path),
            "database_exists": db_file_path.exists(),
            "database_size_mb": 0.0,
            "database_writable": False,
        }
        
        if db_file_path.exists():
            info["database_size_mb"] = db_file_path.stat().st_size / (1024 * 1024)
            info["database_writable"] = os.access(db_file_path, os.W_OK)
        else:
            # Check if directory is writable
            info["database_writable"] = os.access(db_dir_path, os.W_OK)
        
        return info
    except Exception as e:
        logger.error(f"Error getting database info: {e}")
        return {"error": str(e)}


def backup_database(backup_suffix=None):
    """
    Create a backup of the current database.
    
    Args:
        backup_suffix: Optional suffix for backup filename
    
    Returns:
        Path to backup file if successful, None otherwise
    """
    db_file_path = Path(get_database_url().replace("sqlite:///", ""))
    db_dir_path = get_db_dir()
    
    if not db_file_path.exists():
        logger.warning("No database file to backup")
        return None
    
    try:
        if backup_suffix is None:
            from datetime import datetime
            backup_suffix = datetime.now().strftime("%Y%m%d_%H%M%S")
        
        backup_file = db_dir_path / f"vrouter-agent.db.backup_{backup_suffix}"
        
        import shutil
        shutil.copy2(db_file_path, backup_file)
        
        logger.info(f"Database backed up to: {backup_file}")
        return backup_file
        
    except Exception as e:
        logger.error(f"Failed to backup database: {e}")
        return None



"""
StreamHandler module for processing multichain stream data.
Provides enhanced functionality for stream data processing with multichain integration.
"""

import json
import threading
import time
from typing import Optional, Dict, Any, List, Callable
from dataclasses import dataclass
from enum import Enum

from loguru import logger as log
from vrouter_agent.utils.mutichain import (
    get_multichain_client, 
    list_stream_items,
    MultichainConfig,
    multichain_client_context,
    ensure_stream_subscription,
    MultichainConnectionError,
    StreamNotFoundError
)
from vrouter_agent.models import Transaction


class StreamState(Enum):
    """Stream processing state."""
    INITIALIZED = "initialized"
    RUNNING = "running"
    STOPPED = "stopped"
    ERROR = "error"


@dataclass
class StreamConfig:
    """Configuration for stream processing."""
    chain_name: str
    user: str
    stream_name: str
    poll_interval: float = 5.0
    max_retries: int = 3
    timeout: float = 30.0
    batch_size: int = 100


class StreamHandler:
    """
    Enhanced stream handler for processing multichain stream data.
    Integrates with the existing multichain functionality and provides
    robust stream processing capabilities.
    """
    
    def __init__(self, stream: str, txid: str, config: Optional[StreamConfig] = None):
        """
        Initialize the StreamHandler.
        
        Args:
            stream: Stream name to process
            txid: Transaction ID for tracking
            config: Stream configuration (optional)
        """
        self.stream = stream
        self.txid = txid
        self.config = config or StreamConfig(
            chain_name="default",
            user="multichain",
            stream_name=stream
        )
        
        # Create multichain config for enhanced utilities
        self.multichain_config = MultichainConfig(
            chain_name=self.config.chain_name,
            user=self.config.user,
            max_retries=3,
            retry_delay=2.0
        )
        
        self.state = StreamState.INITIALIZED
        self.multichain_client = None
        self.processing_thread = None
        self.stop_event = threading.Event()
        self.error_count = 0
        self.last_processed_txid = None
        self.processed_items = []
        self.callbacks: List[Callable] = []
        
        # Enhanced metrics tracking
        self.metrics = {
            'total_processed': 0,
            'total_errors': 0,
            'start_time': None,
            'last_activity': None,
            'processing_times': [],
            'throughput_per_minute': 0.0
        }
        
        # Transaction tracking
        self.transaction = None
        
        log.info(f"StreamHandler initialized for stream: {stream}, txid: {txid}")
    
    def add_callback(self, callback: Callable) -> None:
        """Add a callback function to be called when data is processed."""
        self.callbacks.append(callback)
        log.debug(f"Added callback: {callback.__name__}")
    
    def start(self) -> bool:
        """
        Start the stream processing.
        
        Returns:
            bool: True if started successfully, False otherwise
        """
        try:
            log.info(f"Starting stream handler for {self.stream}")
            
            # Initialize multichain client
            self.multichain_client = get_multichain_client(
                self.config.chain_name, 
                self.config.user
            )
            
            if not self.multichain_client:
                log.error("Failed to initialize multichain client")
                self.state = StreamState.ERROR
                return False
            
            # Create transaction record
            self._create_transaction()
            
            # Subscribe to stream if not already subscribed
            self._ensure_stream_subscription()
            
            # Start processing thread
            self.processing_thread = threading.Thread(
                target=self._processing_loop,
                daemon=True
            )
            self.processing_thread.start()
            
            self.state = StreamState.RUNNING
            self.metrics['start_time'] = time.time()
            self.metrics['last_activity'] = time.time()
            log.info(f"Stream handler started successfully for {self.stream}")
            return True
            
        except Exception as e:
            log.error(f"Failed to start stream handler: {e}")
            self.state = StreamState.ERROR
            return False
    
    def stop(self) -> None:
        """Stop the stream processing."""
        log.info(f"Stopping stream handler for {self.stream}")
        self.stop_event.set()
        self.state = StreamState.STOPPED
        
        if self.processing_thread and self.processing_thread.is_alive():
            self.processing_thread.join(timeout=5.0)
    
    def read_data(self) -> Optional[Dict[str, Any]]:
        """
        Read data from the stream.
        
        Returns:
            Optional[Dict[str, Any]]: Stream data if available, None otherwise
        """
        if self.state != StreamState.RUNNING:
            return None
        
        try:
            # Get latest processed item
            if self.processed_items:
                return self.processed_items.pop(0)
            
            return None
            
        except Exception as e:
            log.error(f"Error reading data: {e}")
            self.handle_error(e)
            return None
    
    def process_data(self, data: Dict[str, Any], *args, **kwargs) -> bool:
        """
        Process stream data with enhanced metrics tracking.
        
        Args:
            data: Data to process
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
            
        Returns:
            bool: True if processed successfully, False otherwise
        """
        process_start_time = time.time()
        
        try:
            log.debug(f"Processing data: {data}")
            
            # Update metrics
            self.metrics['last_activity'] = process_start_time
            
            # Update transaction
            self._update_transaction_progress(data)
            
            # Call registered callbacks
            for callback in self.callbacks:
                try:
                    callback(data, *args, **kwargs)
                except Exception as e:
                    log.error(f"Callback {callback.__name__} failed: {e}")
                    self.metrics['total_errors'] += 1
            
            # Update last processed txid
            if 'txid' in data:
                self.last_processed_txid = data['txid']
            
            # Update metrics
            process_time = time.time() - process_start_time
            self.metrics['processing_times'].append(process_time)
            self.metrics['total_processed'] += 1
            
            # Keep only last 100 processing times for average calculation
            if len(self.metrics['processing_times']) > 100:
                self.metrics['processing_times'] = self.metrics['processing_times'][-100:]
            
            # Calculate throughput
            self._update_throughput()
            
            log.debug(f"Data processed successfully in {process_time:.3f}s")
            return True
            
        except Exception as e:
            log.error(f"Error processing data: {e}")
            self.metrics['total_errors'] += 1
            self.handle_error(e)
            return False
    
    def _update_throughput(self) -> None:
        """Update throughput metrics."""
        if self.metrics['start_time']:
            elapsed_minutes = (time.time() - self.metrics['start_time']) / 60.0
            if elapsed_minutes > 0:
                self.metrics['throughput_per_minute'] = self.metrics['total_processed'] / elapsed_minutes
    
    def handle_error(self, error: Exception) -> None:
        """
        Handle errors during processing.
        
        Args:
            error: The error that occurred
        """
        self.error_count += 1
        log.error(f"Stream handler error #{self.error_count}: {error}")
        
        if self.error_count >= self.config.max_retries:
            log.error(f"Max retries ({self.config.max_retries}) exceeded, stopping handler")
            self.state = StreamState.ERROR
            self.stop()
        else:
            # Implement exponential backoff
            backoff_time = min(30.0, 2.0 ** self.error_count)
            log.info(f"Retrying in {backoff_time} seconds...")
            time.sleep(backoff_time)
    
    def cleanup(self) -> None:
        """Cleanup resources."""
        log.info(f"Cleaning up stream handler for {self.stream}")
        
        self.stop()
        
        # Final transaction update
        if self.transaction:
            self._finalize_transaction()
        
        # Clear callbacks
        self.callbacks.clear()
        
        # Close multichain client connection if needed
        self.multichain_client = None
        
        log.info("Stream handler cleanup completed")
    
    def get_status(self) -> Dict[str, Any]:
        """
        Get current handler status with enhanced metrics.
        
        Returns:
            Dict[str, Any]: Status information including metrics
        """
        avg_processing_time = 0.0
        if self.metrics['processing_times']:
            avg_processing_time = sum(self.metrics['processing_times']) / len(self.metrics['processing_times'])
        
        uptime = 0.0
        if self.metrics['start_time']:
            uptime = time.time() - self.metrics['start_time']
        
        return {
            "stream": self.stream,
            "txid": self.txid,
            "state": self.state.value,
            "error_count": self.error_count,
            "last_processed_txid": self.last_processed_txid,
            "processed_count": len(self.processed_items),
            "is_running": self.state == StreamState.RUNNING,
            "config": {
                "chain_name": self.config.chain_name,
                "user": self.config.user,
                "stream_name": self.config.stream_name,
                "poll_interval": self.config.poll_interval
            },
            "metrics": {
                "total_processed": self.metrics['total_processed'],
                "total_errors": self.metrics['total_errors'],
                "uptime_seconds": uptime,
                "throughput_per_minute": self.metrics['throughput_per_minute'],
                "avg_processing_time": avg_processing_time,
                "last_activity": self.metrics['last_activity']
            }
        }
    
    # Private methods
    
    def _create_transaction(self) -> None:
        """Create a transaction record for tracking."""
        try:
            self.transaction = Transaction(
                txid=self.txid,
                stream=self.stream
            )
            log.debug(f"Created transaction record: {self.transaction}")
        except Exception as e:
            log.warning(f"Failed to create transaction record: {e}")
    
    def _update_transaction_progress(self, data: Dict[str, Any]) -> None:
        """Update transaction with processing progress."""
        if self.transaction:
            # Update transaction with current processing info
            # This could be extended to store more detailed progress
            pass
    
    def _finalize_transaction(self) -> None:
        """Finalize the transaction record."""
        if self.transaction:
            log.debug(f"Finalizing transaction: {self.transaction}")
            # Here you could save the transaction to a database
            # or perform any final transaction-related operations
    
    def _ensure_stream_subscription(self) -> None:
        """Ensure the stream is subscribed using enhanced utilities."""
        try:
            with multichain_client_context(self.multichain_config) as client:
                # Use the enhanced ensure_stream_subscription function
                ensure_stream_subscription(client, self.config.stream_name)
                log.info(f"Ensured subscription to stream: {self.config.stream_name}")
        except Exception as e:
            log.error(f"Failed to ensure stream subscription: {e}")
            raise
    
    def _processing_loop(self) -> None:
        """Main processing loop running in a separate thread."""
        log.info(f"Starting processing loop for stream: {self.stream}")
        
        while not self.stop_event.is_set():
            try:
                # Get processed txids to skip
                skip_txids = [self.last_processed_txid] if self.last_processed_txid else []
                
                # Use enhanced multichain utilities directly
                items = list_stream_items(
                    config=self.multichain_config,
                    stream_name=self.config.stream_name,
                    count=self.config.batch_size,
                    verbose=True
                )
                
                # Filter out already processed items
                new_items = []
                for item in items:
                    if hasattr(item, 'txid') and item.txid not in skip_txids:
                        new_items.append(item.__dict__ if hasattr(item, '__dict__') else item)
                    elif isinstance(item, dict) and item.get('txid') not in skip_txids:
                        new_items.append(item)
                
                # Add items to processing queue
                for item in new_items:
                    self.processed_items.append(item)
                
                if new_items:
                    log.debug(f"Retrieved {len(new_items)} new items from stream")
                    # Update last processed txid
                    if new_items:
                        last_item = new_items[-1]
                        if hasattr(last_item, 'txid'):
                            self.last_processed_txid = last_item.txid
                        elif isinstance(last_item, dict) and 'txid' in last_item:
                            self.last_processed_txid = last_item['txid']
                
                # Reset error count on successful processing
                if new_items and self.error_count > 0:
                    self.error_count = 0
                
                # Wait before next poll
                self.stop_event.wait(self.config.poll_interval)
                
            except (MultichainConnectionError, StreamNotFoundError) as e:
                log.error(f"Multichain error in processing loop: {e}")
                self.handle_error(e)
                
                if self.state == StreamState.ERROR:
                    break
                    
            except Exception as e:
                log.error(f"Error in processing loop: {e}")
                self.handle_error(e)
                
                if self.state == StreamState.ERROR:
                    break
        
        log.info(f"Processing loop ended for stream: {self.stream}")


def create_enhanced_stream_handler(
    stream: str, 
    txid: str, 
    chain_name: str = "default",
    user: str = "multichain",
    poll_interval: float = 5.0
) -> StreamHandler:
    """
    Factory function to create an enhanced stream handler.
    
    Args:
        stream: Stream name
        txid: Transaction ID
        chain_name: Multichain name
        user: User for multichain connection
        poll_interval: Polling interval in seconds
    
    Returns:
        StreamHandler: Configured stream handler
    """
    config = StreamConfig(
        chain_name=chain_name,
        user=user,
        stream_name=stream,
        poll_interval=poll_interval
    )
    
    return StreamHandler(stream, txid, config)



"""add_topology_id_to_tunnel_config

Revision ID: 1180d19201bd
Revises: acb2cb4a8198
Create Date: 2025-06-08 05:48:59.287378

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '1180d19201bd'
down_revision: Union[str, None] = 'acb2cb4a8198'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # Add topology_id column to tunnel_config_data table
    op.add_column('tunnel_config_data', sa.Column('topology_id', sa.Integer(), nullable=True))


def downgrade() -> None:
    # Remove topology_id column from tunnel_config_data table
    op.drop_column('tunnel_config_data', 'topology_id')



from collections import defaultdict
import configparser
import subprocess
import socket
import os
from cryptography.fernet import Fernet
import re
import tempfile
from pathlib import Path

# import docker
import importlib.metadata
import datetime
from .cli import run_command
from loguru import logger as log


def version():
    """Prints version"""
    return importlib.metadata.version("vrouter-agent")


def get_device_serial_number():
    """
    Returns the serial number of the device.
    """
    val, _ = run_command(["sudo", "dmidecode", "-s", "system-serial-number"])
    return val.strip()

def get_device_short_hostname():
    """
    Returns the hostname of the device.
    """
    val, _ = run_command(["hostname", "-s"])
    return val.strip()

def get_vpp_version():
    """
    Returns the version of VPP.
    """
    val, _ = run_command(["sudo","vppctl", "sh","version"])
    return val.strip()

def get_tunnel_in_scope(order, device_serial_number):
    """
    If order status does not have `requested` status, it is not in scope.
    If the device hostname does not match the order hostname, it is not in scope.

    Params:
        order: Order object.
        device_serial_number: Serial number of the device.

    Returns:
        List of tunnels in scope or empty list.
    """
    tunnels = []
   
    for tunnel in order.tunnels:
        if tunnel.source.device_hostname == device_serial_number:
            tunnels.append(tunnel)
        if tunnel.destination.device_hostname == device_serial_number:
            tunnels.append(tunnel)
    return tunnels


def check_vpp():
    """Check vpp status"""
    try:
        run_command(["sudo", "systemctl", "status", "vpp", "--quiet"])
    except Exception:
        log.error("VPP is not installed or running.")
        return False
    return True


def check_vrouter():
    """Check vrouter status"""
    try:
        run_command(["sudo", "systemctl", "status", "vrouter", "--quiet"])
    except Exception:
        log.error("Vrouter is not installed or running.")
        return False
    return True


def check_frr(frr_container_name: str = "frr"):
    """Check frr status"""
    try:
        run_command(["docker", "inspect", frr_container_name])
    except Exception:
        log.error("FRR is not installed or running. Exiting...")
        return False
    return True


def restart_vpp():
    """Restart vpp"""
    try:
        run_command(["sudo", "systemctl", "restart", "vpp"])
    except Exception:
        log.error("Failed to restart VPP.")
        return False
    return True


def restart_vrouter():
    """Restart vrouter"""
    try:
        run_command(["sudo", "systemctl", "restart", "vrouter"])
    except Exception:
        log.error("Failed to restart vrouter.")
        return False
    return True


def restart_frr(frr_container_name: str = "frr"):
    """Restart frr"""
    try:
        run_command(["docker", "restart", frr_container_name])
    except Exception:
        log.error("Failed to restart FRR.")
        return False
    return True


def restart_services():
    """Restart all services"""
    if not restart_vpp():
        return False
    if not restart_vrouter():
        return False
    if not restart_frr():
        return False
    return True


def check_dependencies():
    log.info("Checking prerequisites...")
    vpp = check_vpp()
    vrouter = check_vrouter()
    frr = check_frr()
    if vpp and vrouter and frr:
        return True
    else:
        return False


def read_config(file_path):
    obj = {}
    config = configparser.ConfigParser()
    config.read(file_path)
    for section in config.sections():
        for key in config[section]:
            obj[key] = config[section][key]
    return obj


# def get_docker_ip_address(container_name: str):
#    """Get the IP address of a docker container

#    Args:
#        container_name (str): Name of the container

#    Returns:
#        str: IP address of the container
#    """

#    client = docker.from_env()
#    container = client.containers.get(container_name)
#    return container.attrs["NetworkSettings"]["IPAddress"]


# def run_docker_command(container_name: str, command: str):
#    """Run a command in a docker container

#    Args:
#        container_name (str): Name of the container
#        command (str): Command to run

#    Returns:
#        str: Output of the command
#    """

#    client = docker.from_env()
#    container = client.containers.get(container_name)
#    return container.exec_run(command)


def get_ip4_addr():
    """Get the IP address of the current machine

    Returns:
        str: IP address of the machine
    """
    return [
        (s.connect(("8.8.8.8", 53)), s.getsockname()[0], s.close())
        for s in [socket.socket(socket.AF_INET, socket.SOCK_DGRAM)]
    ][0][1]


def get_targeted_list(conf_list, conf_dir):
    """Get target list

    Args:
        conf_list (list): list of config file
        conf_dir (str): config dir

    Returns:
        _type_: _description_
    """
    if len(conf_list) == 0:
        log.warning("No configuration files found.")
    else:
        for x in conf_list:
            my_list = {
                "hostname": x["hostname"],
                "tunnel_id": x["tunnel_id"],
                "files": [],
            }
            my_list["files"] += [
                each
                for each in os.listdir(conf_dir)
                if each.startswith("tnn%s" % x["id"])
            ]
            if my_list["hostname"] == socket.gethostname():
                return my_list
    return None


def initialize_symmetric_key(key):
    f = Fernet(key)
    return f


def decrypt_data_symmetric(data, secret_key):
    f = initialize_symmetric_key(secret_key)
    token = f.decrypt(data)
    return token.decode("utf-8")


def decrypt_stream_item(args):
    """Decrypt stream item"""
    log.info("Decrypting stream item...")
    hex_data = args.hex_data
    # convert hex to bytes
    hex_data = bytes.fromhex(hex_data)
    # decrypt data
    log.debug("Decrypting data...")
    order_json = decrypt_data_symmetric(hex_data, args.secret_key)
    log.debug("Data decrypted!")
    return order_json


def is_system_rebooted():
    try:
        result = subprocess.check_output(["uptime", "-s"])
        boot_time = datetime.datetime.strptime(
            result.decode("utf-8").strip(), "%Y-%m-%d %H:%M:%S"
        )
        now = datetime.datetime.now()
        if (
            now - boot_time
        ).total_seconds() < 60:  # Check if the server has been rebooted in the last minute
            return True
        else:
            return False
    except Exception as e:
        print(f"An error occurred: {e}")
        return False


def convert_wg_to_gre_ip_addr(ip_address):
    """
    Manipulate the ip address to be in the range of 192.168.20x.0/24

    For example: 192.168.101.10 -> 192.168.201.10

    """
    ip_parts = ip_address.split(".")
    third_octet = ip_parts[2]
    if len(third_octet) >= 3:
        # If third octet has 3+ digits, take the last digit and prepend "20"
        ip_parts[2] = "20" + third_octet[2:]
    else:
        # If third octet has 1-2 digits, prepend "20" to the whole octet
        ip_parts[2] = "20" + third_octet
    return ".".join(ip_parts)


def convert_subnet_to_mask(subnet):
    return str(sum([bin(int(x)).count("1") for x in subnet.split(".")]))


def get_primary_interface(interfaces):
    """Filters the given interfaces to only include the primary one.

    Args:
      interfaces: A list of dictionaries representing interfaces,
        each with an 'is_primary' key.

    Returns:
      A list containing only the primary interface, or an empty list if
      no primary interface is found.
    """

    # Filter interfaces based on the is_primary flag
    primary_interfaces = [
        interface for interface in interfaces if interface["is_primary"] == "True"
    ]

    # Return the first primary interface (assuming only one primary exists)
    return primary_interfaces[0] if primary_interfaces else None


def extract_wan_ip_addresses(input_str):
    # Pattern to match IP addresses with labels
    pattern = (
        r"(?P<ip>\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})\((?P<label>primary|secondary)\)"
    )

    matches = re.findall(pattern, input_str)

    ip_dict = {"primary": None, "secondary": None}

    if matches:
        # Populate the dictionary with labeled IP addresses
        for ip, label in matches:
            ip_dict[label] = ip
    else:
        # If no labeled IPs are found, return the single IP address with key "primary"
        ip_pattern = r"\b\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}\b"
        match = re.search(ip_pattern, input_str)
        if match:
            ip_dict["primary"] = match.group(0)

    return ip_dict


def get_temp_dir():
    """Get a secure temporary directory for the application."""
    temp_dir = Path(tempfile.gettempdir()) / "vrouter-agent"
    temp_dir.mkdir(exist_ok=True, mode=0o700)  # Secure permissions
    return temp_dir

def get_frr_config_path():
    """Get the path for FRR configuration file."""
    return get_temp_dir() / "frr.conf"

def get_frr_sync_dir():
    """Get the directory for FRR sync files."""
    return get_temp_dir() / "frr-sync"



"""
Pydantic schemas for tunnel configuration data.
These schemas match the tunnel_data structure from the Django tasks.py file.
"""

from typing import List, Dict, Any, Optional, Union
from pydantic import BaseModel, Field, validator
from enum import Enum


class TunnelState(str, Enum):
    """Enumeration of possible tunnel configuration states"""
    PLANNED = "planned"
    ACTIVE = "active"
    INACTIVE = "inactive"
    ERROR = "error"
    PENDING = "pending"


class TopologyInfo(BaseModel):
    """Topology information in tunnel configuration"""

    id: Union[str, int] = Field(description="Topology ID")
    name: str = Field(description="Topology name")
    type: str = Field(
        description="Topology type (hub_spoke, point_to_point, full_mesh)"
    )
    
    @validator('id', pre=True)
    def convert_id_to_string(cls, v):
        """Convert topology ID to string if it's an integer"""
        return str(v) if v is not None else None


class ClientInterface(BaseModel):
    """Client interface configuration for VPP and routing"""
    
    name: str = Field(description="Interface name (e.g., vpp-lan)")
    type: str = Field(description="Interface type (e.g., lan, wan)")
    ip_address: str = Field(description="IP address of the interface")
    subnet_mask: int = Field(description="Subnet mask (CIDR notation)")
    mac_address: Optional[str] = Field(None, description="MAC address of the interface")
    vpp_whitelisted: Optional[bool] = Field(None, description="Whether interface is whitelisted in VPP")
    distribution_mode: Optional[str] = Field(None, description="Routing distribution mode (ospf, bgp)")
    ospf_area: Optional[str] = Field(None, description="OSPF area for this interface")
    ospf_cost: Optional[int] = Field(None, description="OSPF cost for this interface")
    bgp_remote_asn: Optional[int] = Field(None, description="Remote BGP ASN if using BGP")
    is_enabled: Optional[bool] = Field(True, description="Whether the interface is enabled")

class BgpPeer(BaseModel):
    """BGP peer configuration for VPP and routing"""
    
    remote_asn: int = Field(description="Remote ASN for this BGP peer")
    peer_ip: str = Field(description="IP address of the BGP peer")
    peer_type: str = Field(description="Type of BGP peer (internal, external)")
    multihop: Optional[int] = Field(0, description="Whether multihop is enabled")
    next_hop_self: Optional[bool] = Field(False, description="Whether next-hop self is enabled")
    is_enabled: Optional[bool] = Field(True, description="Whether the BGP peer is enabled")
    lan_interface_ip: Optional[str] = Field(None, description="LAN interface IP address for this peer")
    lan_mac_address: Optional[str] = Field(None, description="LAN MAC address for this peer")

    @validator('remote_asn')
    def convert_asn_to_int(cls, v):
        """Convert ASN to integer if it's a string"""
        return int(v) if isinstance(v, str) and v.isdigit() else v
    
    
class WireGuardConfig(BaseModel):
    """WireGuard tunnel configuration"""

    interface_name: str = Field(description="WireGuard interface name")
    private_key: Optional[str] = Field(None, description="Private key for this node")
    listen_port: Optional[int] = Field(None, description="Listen port for WireGuard")
    address: Optional[str] = Field(None, description="Tunnel IP address")
    peer_public_key: Optional[str] = Field(None, description="Peer's public key")
    allowed_ips: Optional[str] = Field(None, description="Allowed IPs for the peer")
    persistent_keepalive: int = Field(default=25, description="Keepalive interval")
    peer_endpoint: Optional[str] = Field(None, description="Peer endpoint IP:port")
    mtu: Optional[int] = Field(default=1420, description="Maximum transmission unit")
    peer_address: Optional[str] = Field(None, description="Peer IP address for routing")
    source_ip: Optional[str] = Field(None, description="Source IP address for tunnel")
    
    # VPP Interface Data Fields (embedded from VPP interface mappings)
    vpp_interface_name: Optional[str] = Field(None, description="VPP interface name (e.g., wg0, wg1)")
    vpp_ip_address: Optional[str] = Field(None, description="VPP interface IP address")
    vpp_subnet_mask: Optional[int] = Field(None, description="VPP interface subnet mask")
    vpp_type: Optional[str] = Field(None, description="VPP interface type (tunnel)")
    vpp_used: Optional[bool] = Field(None, description="Whether VPP interface is in use")
    vpp_interface_index: Optional[int] = Field(None, description="VPP interface index")
    vpp_status: Optional[str] = Field(None, description="VPP interface status (created, up, down)")
    vpp_up: Optional[bool] = Field(None, description="Whether VPP interface is up")
    vpp_operational: Optional[bool] = Field(None, description="Whether VPP interface is operational")
    vpp_connectivity_test_passed: Optional[bool] = Field(None, description="Whether connectivity test passed")
    vpp_created_at: Optional[str] = Field(None, description="Timestamp when VPP interface was created")
    vpp_last_verified_at: Optional[str] = Field(None, description="Timestamp of last verification")
    
    # Tunnel State Management
    state: Optional[TunnelState] = Field(None, description="Current operational state of the tunnel")
    tunnel_id: Optional[str] = Field(None, description="Unique tunnel identifier")
    action: Optional[str] = Field(None, description="Action to perform on this tunnel (create, update, delete)")
    
    # GRE Tunnel Configuration
    gre_local_name: Optional[str] = Field(None, description="GRE tunnel local interface name")
    gre_remote_name: Optional[str] = Field(None, description="GRE tunnel remote interface name")
    gre_local_tunnel_ip: Optional[str] = Field(None, description="Local GRE tunnel IP address")
    gre_remote_tunnel_ip: Optional[str] = Field(None, description="Remote GRE tunnel IP address")
    
    @validator('tunnel_id', pre=True)
    def convert_tunnel_id_to_string(cls, v):
        """Convert tunnel_id to string if it's an integer"""
        return str(v) if v is not None else None



class FrrConfig(BaseModel):
    """FRR routing configuration structure"""

    zebra_config: Optional[str] = Field(None, description="Zebra daemon configuration")
    ospf_config: Optional[str] = Field(None, description="OSPF routing configuration")
    bgp_config: Optional[str] = Field(None, description="BGP routing configuration")
    static_config: Optional[str] = Field(None, description="Static routes configuration")
    full_config: Optional[str] = Field(None, description="Complete FRR configuration")
    config_hash: Optional[str] = Field(None, description="Configuration hash for validation")

    def to_config_string(self) -> str:
        """
        Convert the structured FRR configuration to a single configuration string.
        
        Returns:
            str: Complete FRR configuration as a string
        """
        # If full_config is available, use it
        if self.full_config:
            return self.full_config
        
        # Otherwise, build from components
        config_parts = []
        
        # Add header
        config_parts.append("! FRR Configuration")
        config_parts.append("! Generated by VRouter Agent")
        config_parts.append("!")
        
        # Add each configuration section
        if self.zebra_config:
            config_parts.append("! Zebra Configuration")
            config_parts.append(self.zebra_config.strip())
            config_parts.append("!")
        
        if self.ospf_config:
            config_parts.append("! OSPF Configuration")
            config_parts.append(self.ospf_config.strip())
            config_parts.append("!")
        
        if self.bgp_config:
            config_parts.append("! BGP Configuration")
            config_parts.append(self.bgp_config.strip())
            config_parts.append("!")
        
        if self.static_config:
            config_parts.append("! Static Routes Configuration")
            config_parts.append(self.static_config.strip())
            config_parts.append("!")
        
        return "\n".join(config_parts)

    @classmethod
    def from_string(cls, config_string: str) -> "FrrConfig":
        """
        Create an FrrConfig instance from a configuration string.
        This is a basic parser - for complex configurations, you may need a more sophisticated approach.
        
        Args:
            config_string: FRR configuration as a string
            
        Returns:
            FrrConfig: Parsed configuration object
        """
        return cls(full_config=config_string.strip())

    def is_empty(self) -> bool:
        """Check if the configuration is empty."""
        return not any([
            self.zebra_config,
            self.ospf_config, 
            self.bgp_config,
            self.static_config,
            self.full_config
        ])


class TunnelConfigData(BaseModel):
    """Main tunnel configuration data structure"""

    tag: str = Field(description="Configuration tag", default="tunnel_config")
    action: str = Field(description="Action to perform (provision, decommission)")
    order_id: str = Field(description="Order ID")
    order_number: str = Field(description="Order number")
    topology: Optional[TopologyInfo] = Field(None, description="Topology information")
    tunnels: List[WireGuardConfig] = Field(description="List of tunnel configurations")
    client_interfaces: Optional[List[ClientInterface]] = Field(
        None, description="List of client interface configurations"
    )
    bgp_peers: Optional[List[BgpPeer]] = Field(
        None, description="List of BGP peer configurations if eBGP is used"
    )
    ospf_enabled: Optional[bool] = Field(
        default=False, description="Whether OSPF protocol is enabled"
    )
    ebgp_enabled: Optional[bool] = Field(
        default=False, description="Whether eBGP protocol is enabled"
    )
    frr_config: Optional[FrrConfig] = Field(
        None, description="FRR configuration if available"
    )
    state: TunnelState = Field(
        default=TunnelState.PLANNED, 
        description="Current state of the tunnel configuration"
    )
 

class NodeTunnelConfig(BaseModel):
    """Node-specific tunnel configuration for blockchain posting"""

    hostname: str = Field(description="Node hostname")
    config: TunnelConfigData = Field(description="Tunnel configuration for this node")


class BlockchainPostData(BaseModel):
    """Data structure for posting to blockchain"""

    tunnel_data: TunnelConfigData = Field(description="Tunnel configuration data")
    encrypted_data: str = Field(description="Encrypted hex data for blockchain")
    stream_name: str = Field(description="Blockchain stream name")
    txid: Optional[str] = Field(None, description="Transaction ID after posting")


# Request/Response schemas for API endpoints
class TunnelConfigRequest(BaseModel):
    """Request schema for tunnel configuration"""

    order_id: str
    topology_id: str
    action: str = Field(default="provision", pattern="^(provision|decommission)$")


class TunnelConfigResponse(BaseModel):
    """Response schema for tunnel configuration"""

    success: bool
    message: str
    data: Optional[Dict[str, TunnelConfigData]] = None
    errors: Optional[List[str]] = None


class BlockchainPostRequest(BaseModel):
    """Request schema for blockchain posting"""

    order_id: str
    node_hostname: Optional[str] = Field(
        None, description="Specific node to post config for"
    )


class BlockchainPostResponse(BaseModel):
    """Response schema for blockchain posting"""

    success: bool
    results: List[Dict[str, Any]] = Field(description="Results per node")
    errors: Optional[List[str]] = None



import ipaddress
import os
from itertools import chain
from typing import List
from loguru import logger as log
from vpp_vrouter.common import models

from vrouter_agent.utils import run_command

WG_TUNNELS_PER_CONNECTION = 4


def add_vpp_route(
    client,
    dest_net: str,
    next_hop_addr: str,
    outgoing_iface: str,
):
    reply = client.add_configuration(
        models.RouteConfigurationItem(
            destination_network=dest_net,
            next_hop_address=next_hop_addr,
            outgoing_interface=outgoing_iface,
        )
    )
    log.debug(reply)  # TODO error handling


def add_linux_route(
    client,
    dest_net: str,
    next_hop_addr: str = "",
    outgoing_iface: str = "",
    container: str = "",  # empty container name means default linux namespace
):
    reply = client.add_configuration(
        models.LinuxRouteConfigurationItem(
            docker_container_name=container,
            destination_network=dest_net,
            next_hop_address=next_hop_addr,
            outgoing_interface=outgoing_iface,
        )
    )
    log.debug(reply)  # TODO error handling


def add_lcp_interface(
    client,
    vpp_iface: str,
    host_iface: str,
    host_netns: str,
    is_tun: bool = False,
):
    iface_type = (
        models.LCPHostInterfaceTypeEnum.TUN
        if is_tun
        else models.LCPHostInterfaceTypeEnum.TAP
    )
    reply = client.add_configuration(
        models.LCPPairConfigurationItem(
            interface=vpp_iface,
            mirror_interface_host_name=host_iface,
            mirror_interface_type=iface_type,
            host_namespace=host_netns,
        )
    )
    log.debug(reply)  # TODO error handling


def add_wireguard_interface(
    client,
    interface_name: str,
    interface_ip_network: str,
    port: int,
    private_key: str,
    src_ip: str,
    mtu: int,
):
    reply = client.add_configuration(
        models.InterfaceConfigurationItem(
            name=interface_name,
            type=models.InterfaceType.WIREGUARD_TUNNEL,
            enabled=True,
            ip_addresses=[interface_ip_network],
            mtu=mtu,
            link=models.WireguardInterfaceLink(
                private_key=private_key,
                port=port,
                src_addr=ipaddress.IPv4Address(src_ip),
            ),
        )
    )
    log.debug(reply)  # TODO error handling


def add_wireguard_peer(
    client,
    interface_name: str,
    port: int,
    endpoint: str,
    remote_public_key: str,
    persistent_keepalive: int,
):
    reply = client.add_configuration(
        models.WireguardPeerConfigurationItem(
            public_key=remote_public_key,
            port=port,
            endpoint=ipaddress.IPv4Address(endpoint),
            allowed_ips=["0.0.0.0/0"],
            wg_if_name=interface_name,
            persistent_keepalive=persistent_keepalive,
        )
    )
    log.debug(reply)  # TODO error handling


def add_wireguard_tunnel(
    client,
    interface_name: str,
    interface_ip: str,
    interface_ip_network: str,
    port: int,
    private_key: str,
    remote_public_key: str,
    remote_ip_network: str,
    remote_tunnel_end_ip: str,
    gre_interface_name: str,
    gre_interface_ip_address: str,
    gre_interface_ip_address_mask: str,
    remote_gre_interface_ip_address: str,
    src_ip: str,
    dst_ip: str,
    mtu: int,
    keepalive: int,
):
    add_wireguard_interface(
        client,
        interface_name,
        interface_ip_network,
        port,
        private_key,
        src_ip,
        mtu,
    )
    add_wireguard_peer(
        client,
        interface_name,
        port,
        dst_ip,
        remote_public_key,
        keepalive,
    )

    # add_route_to_wireguard_peer(interface_name, remote_ip_network, remote_tunnel_end_ip)
    add_gre_tunnel_to_wg_and_lcp_it_to_frr_container(
        client,
        gre_interface_name,
        gre_interface_ip_address,
        gre_interface_ip_address_mask,
        remote_gre_interface_ip_address,
        interface_name,
        remote_tunnel_end_ip,
    )

    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            remote_gre_interface_ip_address,
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            gre_interface_name,
            "Accept",
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            remote_gre_interface_ip_address,
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            "local",  # static forwarding to local processing (it will be processed by LCP)
            "Forward",
        ]
    )

    add_frr_agent_ip_translation_from_gre_to_wg(
        gre_interface_ip_address,
        interface_ip,
        remote_gre_interface_ip_address,
        remote_tunnel_end_ip,
    )
    return interface_name, f"{gre_interface_ip_address}/{gre_interface_ip_address_mask}"


def add_frr_agent_ip_translation_from_gre_to_wg(
    gre_interface_ip_address: str,
    interface_ip: str,
    remote_gre_interface_ip_address: str,
    remote_tunnel_end_ip: str,
    frr_container_name: str = "frr",
    replace_file_path: str = "/rr-ip-replace.txt",
):
    # TODO when proper python client API will be created for this then replace this with python client code
    _, code1 = run_command(
        [
            "docker",
            "exec",
            "-t",
            frr_container_name,
            "sh",
            "-c",
            "echo '"
            + gre_interface_ip_address
            + "=>"
            + interface_ip
            + "' >> "
            + replace_file_path,
        ]
    )
    _, code2 = run_command(
        [
            "docker",
            "exec",
            "-t",
            frr_container_name,
            "sh",
            "-c",
            "echo '"
            + remote_gre_interface_ip_address
            + "=>"
            + remote_tunnel_end_ip
            + "' >> "
            + replace_file_path,
        ]
    )
    return code1 == 0 and code2 == 0


def add_bgp_loopback(client, ip_addr: str):
    reply = client.add_configuration(
        models.InterfaceConfigurationItem(
            name="loop0",
            type=models.InterfaceType.SOFTWARE_LOOPBACK,
            enabled=True,
            ip_addresses=[ip_addr + "/32"],
        )
    )
    log.debug(reply)  # TODO error handling


def add_gre_tunnel_to_wg_and_lcp_it_to_frr_container(
    client,
    gre_interface_name: str,
    gre_interface_ip_address: str,
    gre_interface_ip_address_mask: str,
    remote_gre_interface_ip_address: str,
    interface_name: str,
    remote_tunnel_end_ip: str,
    frr_container_name: str = "frr",
):
    gre_iface = models.InterfaceConfigurationItem(
        name=gre_interface_name,
        type=models.InterfaceType.GRE_TUNNEL,
        enabled=False,
        ip_addresses=[],
        link=models.GREInterfaceLink(
            type=models.GRELinkType.L3,
            src_addr=ipaddress.IPv4Address(gre_interface_ip_address),
            dst_addr=ipaddress.IPv4Address(remote_gre_interface_ip_address),
        ),
    )

    reply = client.add_configuration(gre_iface)
    log.debug(reply)  # TODO error handling

    add_vpp_route(
        client,
        remote_gre_interface_ip_address
        + "/32",  # route to one ip address => mask /32 is static value
        remote_tunnel_end_ip,
        interface_name,
    )

    add_lcp_interface(
        client,
        gre_interface_name,
        interface_name,
        frr_container_name,
        True,
    )

    reply = client.update_configuration(
        gre_iface,
        models.InterfaceConfigurationItem(
            name=gre_interface_name,
            type=models.InterfaceType.GRE_TUNNEL,
            enabled=True,
            ip_addresses=[
                gre_interface_ip_address + "/" + gre_interface_ip_address_mask
            ],
            link=models.GREInterfaceLink(
                type=models.GRELinkType.L3,
                src_addr=ipaddress.IPv4Address(gre_interface_ip_address),
                dst_addr=ipaddress.IPv4Address(remote_gre_interface_ip_address),
            ),
        ),
    )
    log.debug(reply)  # TODO error handling


def add_frr_configuration(
    client,
    wg_iface_info: dict[str, str],
    frr_hostname: str,
    bgp_loop_ip_addr: str,
    rr_bgp_loop_ip_addr: str,
    customer_iface: str,
    customer_ip_addr: str,
    customer_ip_mask: str,
    customer_ip_net: str,
    customer_ospf_area: str,
    customer_ospf_router_id: str,
    core_ospf_router_id: str,
    customer_ospf_hello_interval: int = 2,
    customer_ospf_dead_interval: int = 8,
    core_ospf_hello_interval: int = 2,
    core_ospf_dead_interval: int = 8,
    core_ospf_cost: int = 10,
    bgp_connection_retry_interval: int = 10,
    customer_ospf_filtering: str = "",
    ospf_zebra_filtering: str = "",
):
    ospf_bgp_redistribution_route_map_ref = (
        " route-map customer-ospf-filtering" if customer_ospf_filtering else ""
    )
    ospf_bgp_redistribution_route_map = (
        f"""{customer_ospf_filtering}
ip prefix-list other seq 5 permit any
!
route-map customer-ospf-filtering deny 5
 match ip address prefix-list customer-underlay
!
route-map customer-ospf-filtering permit 10
 match ip address prefix-list other
!"""
        if customer_ospf_filtering
        else ""
    )
    ospf_zebra_route_map = (
        f"""
!
! filtering out installation of WG and BGP loopback subnets that are not needed for this node
ip protocol ospf route-map ospf-zebra-filtering
!
{ospf_zebra_filtering}
ip prefix-list other2 seq 5 permit any
!
route-map ospf-zebra-filtering permit 5
match ip address prefix-list wg-and-bgp-loopbacks-needed
!
route-map ospf-zebra-filtering deny 10
 match ip address prefix-list all-wg-and-bgp-loopbacks
!
route-map ospf-zebra-filtering permit 15
 match ip address prefix-list other2
!"""
        if ospf_zebra_filtering
        else ""
    )

    wg_interface_frr_config = ""
    for wg_name, wg_ip in wg_iface_info.items():
        wg_interface_frr_config += f"""!
interface {wg_name}
 ip address {wg_ip}
 # 0.0.0.0 is static as all core-OSPF ends on all nodes are in the same OSPF area
 ip ospf 2 area 0.0.0.0
 ip ospf network point-to-point
 ip ospf hello-interval {core_ospf_hello_interval}
 ip ospf dead-interval {core_ospf_dead_interval}
 ip ospf cost {core_ospf_cost}
!
"""
    config_str = f"""!
frr version 8.3
frr defaults traditional
hostname {frr_hostname}
service integrated-vtysh-config
log file /var/log/frr/debug.log
log stdout
log commands
no ipv6 forwarding
!
debug zebra events
debug bgp zebra
debug bgp neighbor-events
debug bgp updates
debug bgp keepalives
debug bgp update-groups
debug bgp nht
debug ospf event
debug ospf zebra
#debug ospf packet all send detail
#debug ospf packet all recv detail
debug static events
debug static route
!
interface lo
 ip address {bgp_loop_ip_addr}/32
 ip ospf 2 area 0.0.0.0
!
{wg_interface_frr_config}
interface {customer_iface}
 ip address {customer_ip_addr}/{customer_ip_mask}
 ip ospf 1 area {customer_ospf_area}
 ip ospf network point-to-point
 ip ospf hello-interval {customer_ospf_hello_interval}
 ip ospf dead-interval {customer_ospf_dead_interval}
!
router ospf 1
 ospf router-id {customer_ospf_router_id}
 redistribute bgp{ospf_bgp_redistribution_route_map_ref}
!
router ospf 2
 ospf router-id {core_ospf_router_id}
!
# BGP ASN is the same for all BGP nodes -> 64512 is static ASN
router bgp 64512
 bgp router-id {bgp_loop_ip_addr}
# no bgp ebgp-requires-policy
# no bgp suppress-duplicates
# no bgp network import-check
 neighbor {rr_bgp_loop_ip_addr} remote-as internal
 neighbor {rr_bgp_loop_ip_addr} update-source lo
 # in case of peer connection failure, retry connection every 10s
 neighbor {rr_bgp_loop_ip_addr} timers connect {bgp_connection_retry_interval}
# neighbor {rr_bgp_loop_ip_addr} timers 5 15
 !
 address-family ipv4 unicast
  network {customer_ip_net}
  redistribute ospf 1
 exit-address-family
!{ospf_bgp_redistribution_route_map}{ospf_zebra_route_map}
line vty
!
"""
    # Apply FRR configuration to the system
    reply = client.add_configuration(models.FRRConfigurationItem(config=config_str))
    log.debug(reply)  # TODO error handling


def add_nat_configuration(
    client,
    lan_iface: str,
    wan_iface: str,
    wan_iface_addr: str,
    wg_ports_to_open: List[int],
    tcp_ports: List[int] = [],
    udp_ports: List[int] = [],
):
    lan_nat_iface = models.Nat44InterfaceConfigurationItem(
        name=lan_iface,
        nat_inside=True,
        nat_outside=False,
        output_feature=False,
    )
    wan_nat_iface = models.Nat44InterfaceConfigurationItem(
        name=wan_iface,
        nat_inside=False,
        nat_outside=True,
        output_feature=False,
    )
    nat_address_pool = models.Nat44AddressPoolConfigurationItem(
        name="nat-pool",
        first_ip=ipaddress.IPv4Address(wan_iface_addr),
        last_ip=ipaddress.IPv4Address(wan_iface_addr),
    )

    wg_port_mappings = [
        models.IdentityMapping(
            interface=wan_iface, protocol=models.ProtocolInNAT.UDP, port=p
        )
        for p in wg_ports_to_open
    ]
    tcp_port_mappings = [
        models.IdentityMapping(
            interface=wan_iface, protocol=models.ProtocolInNAT.TCP, port=p
        )
        for p in tcp_ports
    ]
    udp_port_mappings = [
        models.IdentityMapping(
            interface=wan_iface, protocol=models.ProtocolInNAT.UDP, port=p
        )
        for p in udp_ports
    ]

    nat_mappings = models.DNat44ConfigurationItem(
        label="nat-mappings",
        static_mappings=[],
        identity_mappings=list(
            chain(wg_port_mappings, tcp_port_mappings, udp_port_mappings)
        ),
    )

    reply = client.add_configuration(
        lan_nat_iface,
        wan_nat_iface,
        nat_address_pool,
        nat_mappings,
    )
    log.debug(reply)  # TODO error handling


def add_acl_configuration(
    client,
    ingress_iface: str,
    deny_network: str,
):
    reply = client.add_configuration(
        models.ACLConfigurationItem(
            name="lab-acl-rules",
            ingress=[ingress_iface],
            rules=[
                models.ACLRuleConfigurationItem(
                    action=models.ACLAction.DENY,
                    refinement=models.IPSpecification(
                        addresses=models.IPAddresses(
                            destination_network=ipaddress.IPv4Network(deny_network)
                        )
                    ),
                ),
                models.ACLRuleConfigurationItem(
                    action=models.ACLAction.PERMIT,
                ),
            ],
        )
    )
    log.debug(reply)  # TODO error handling


def add_lcp_global_configuration(client, network_namespace: str):
    reply = client.add_configuration(
        models.LCPGlobalsConfigurationItem(
            default_namespace=network_namespace,
            lcp_sync=True,
            lcp_auto_subint=True,
        )
    )
    log.debug(reply)  # TODO error handling


#
def add_node_part_of_client_to_node_frr_path(client, use_ebgp):
    if use_ebgp:
        add_ebgp_route_in_frr_container(client)
        add_ebgp_routes_in_vpp(client)
    else:
        add_ospf_lcp_fix_env(client)


def add_ebgp_route_in_frr_container(client):
    # direction to client
    add_linux_route(
        client,
        os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"],
        "",
        os.environ[os.environ["VPP"] + "_LAN_INTF"],
        os.environ["NODE_FRR_CONTAINER_NAME"],
    )
    # note: the opposite direction is handled automatically
    #  by linux network namespace as it points to present loopback interface


def add_ebgp_routes_in_vpp(client):
    # direction to client
    add_vpp_route(
        client,
        os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"] + "/32",
        os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
        os.environ[os.environ["VPP"] + "_LAN_INTF"],
    )
    # note: the opposite direction is handled correctly by LCP

    # fixing VPP's neighbor view of client's BGP loop (despite client responding to ARP packets, VPP keeps sending
    # them and doesn't learn needed things...)
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "neighbor",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_MAC"],
        ]
    )


#
def add_node_part_of_client_to_node_frr_path(client, use_ebgp):
    if use_ebgp:
        add_ebgp_route_in_frr_container(client)
        add_ebgp_routes_in_vpp(client)
    else:
        add_ospf_lcp_fix_env(client)


def add_ebgp_route_in_frr_container(client):
    # direction to client
    add_linux_route(
        client,
        os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"],
        "",
        os.environ[os.environ["VPP"] + "_LAN_INTF"],
        os.environ["NODE_FRR_CONTAINER_NAME"],
    )
    # note: the opposite direction is handled automatically
    #  by linux network namespace as it points to present loopback interface


def add_ebgp_routes_in_vpp(client):
    # direction to client
    add_vpp_route(
        client,
        os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"] + "/32",
        os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
        os.environ[os.environ["VPP"] + "_LAN_INTF"],
    )
    # note: the opposite direction is handled correctly by LCP

    # fixing VPP's neighbor view of client's BGP loop (despite client responding to ARP packets, VPP keeps sending
    # them and doesn't learn needed things...)
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "neighbor",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            os.environ[os.environ["CLIENT"] + "_BGP_LOOP_IP"],
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_MAC"],
        ]
    )


def add_ospf_lcp_fix_env(client):
    # https://lists.fd.io/g/vpp-dev/topic/83103366#19478
    # for 224.0.0.5 and 224.0.0.6 used by OSPF (https://en.wikipedia.org/wiki/Open_Shortest_Path_First)

    # run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", "local", "Forward"])
    # for i in range(4):
    #     run_command(["sudo", "vppctl", "ip", "mroute", "add", "224.0.0.0/24", "via", os.environ[os.environ["VPP"]+f"_WG{i}_INTERFACE_NAME"], "Accept"])

    # to steer OSPF Hello packets from FRR container to client
    add_vpp_route(
        client,
        "224.0.0.0/24",  # static due to OSPF hello packets always destined for 224.0.0.5
        os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
        os.environ[os.environ["VPP"] + "_LAN_INTF"],
    )

    # to get ospf hello packets from client to frr container through vpp
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            os.environ[os.environ["VPP"] + "_LAN_INTF"],
            "Accept",
        ]
    )
    run_command(
        [
            "sudo",
            "vppctl",
            "ip",
            "mroute",
            "add",
            os.environ[os.environ["CLIENT"] + "_VPP_INTF_IP"],
            "224.0.0.5",  # static destination ip address of Hello OSPF packets, fixed by OSPF protocol itself
            "via",
            "local",  # static, set to be locally processed (processed by LCP)
            "Forward",
        ]
    )


# Note: this does not include RR-node wireguard tunnel (for that see makefile) and RR setup => only path from linux WG
# on node to BGP in container on node
def add_routes_for_bidirectional_bgp_communication_env(client):
    # Some path explanation:
    # RR-to-node path: node-linux-wg0 -(adding here route1+iptables forwarding)-> FRR container default interface ->
    #   FRR container routing to local loopback -(ping echo going back, need to add route2)-> FRR container default
    #   interface -(routing path was added as part of wg configuration)-> node-linux-wg0
    # Node-to-RR path: FRR instance in FRR container -(need to add route3)-> FRR container default interface
    #   -> node linux host route(actually controlled by WG allowed ip configuration) -> wg tunnel to RR

    # this can be hardcoded unless you want change default networking for all docker containers
    host_ip_from_docker_container = "172.17.0.1"

    # adding route1
    frr_container_ip_address = run_command(
        [
            "docker",
            "inspect",
            "-f",
            "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}'",
            os.environ["NODE_FRR_CONTAINER_NAME"],
        ]
    )
    frr_container_ip_address = frr_container_ip_address.replace("\n", "").replace(
        "'", ""
    )
    add_linux_route(
        client,
        os.environ["LAB_" + os.environ["NODE"] + "_BGP_LOOP_IP"],
        frr_container_ip_address,
    )
    run_command(
        [
            "sudo",
            "iptables",
            "-A",
            "FORWARD",
            "-i",
            os.environ["LAB_" + os.environ["NODE"] + "_LINUX_WG_INTERFACE"],
            "-o",
            "docker0",  # static for default docker installation
            "-j",
            "ACCEPT",
        ]
    )
    # adding route2
    add_linux_route(
        client,
        os.environ["LAB_RR_WG_INTERFACE_IP"],
        host_ip_from_docker_container,
        "",
        os.environ["NODE_FRR_CONTAINER_NAME"],
    )
    # adding route3
    add_linux_route(
        client,
        os.environ["LAB_RR_BGP_LOOP_IP"],
        host_ip_from_docker_container,
        "",
        os.environ["NODE_FRR_CONTAINER_NAME"],
    )
    # TODO check that setting "echo 1 > /proc/sys/net/ipv4/ip_forward" is always set for node (it seems to be true,
    #  just don't know where it is persisted to survive the machine reboot)



"""
Route Manager

This module provides a specialized manager for route operations,
implementing functionality for both Linux and Docker route management.
"""

from typing import Dict, Any, Optional, List
import ipaddress
from loguru import logger as log

from vpp_vrouter.common import models
from vrouter_agent.utils.cli import run_command
from ..connection.vpp_connection import VPPConnectionManager


class RouteManager:
    """
    Manages route operations for both Linux and Docker environments.
    
    This class handles various routing operations including adding and removing
    routes in Linux and Docker environments.
    """
    
    def __init__(self, connection_manager: VPPConnectionManager):
        """
        Initialize the Route Manager.
        
        Args:
            connection_manager: VPP connection manager instance
        """
        self.connection = connection_manager
        self.client = connection_manager.client
        
    def _extract_reply_errors(self, reply) -> List[str]:
        """Extract all error messages from a VPP reply object."""
        errors = []
        
        # Check processing_error (legacy field)
        if hasattr(reply, 'processing_error') and reply.processing_error:
            errors.append(reply.processing_error)
        
        # Check added_items for errors
        if hasattr(reply, 'added_items'):
            for item in reply.added_items:
                if hasattr(item, 'error') and item.error:
                    errors.append(f"Added item error: {item.error}")
                if hasattr(item, 'state') and 'FAILURE' in str(item.state):
                    errors.append(f"Item state failure: {item.state}")
        
        # Check vpp_apply_attempted_items for detailed errors
        if hasattr(reply, 'vpp_apply_attempted_items'):
            for item in reply.vpp_apply_attempted_items:
                if hasattr(item, 'error') and item.error:
                    errors.append(f"VPP apply error: {item.error}")
                if hasattr(item, 'state') and 'FAILURE' in str(item.state):
                    errors.append(f"VPP apply state failure: {item.state}")
        
        # Check vpp_apply_success flag
        if hasattr(reply, 'vpp_apply_success') and reply.vpp_apply_success is False:
            if not errors:  # Only add generic message if no specific errors found
                errors.append("VPP apply failed without specific error message")
        
        return errors
        
 
    def add_linux_route(
        self,
        dest_net: str,
        next_hop_addr: str = "",
        outgoing_iface: str = "",
        container: str = "",  # empty container name means default linux namespace
    ):
        reply = self.client.add_configuration(
            models.LinuxRouteConfigurationItem(
                docker_container_name=container,
                destination_network=dest_net,
                next_hop_address=next_hop_addr,
                outgoing_interface=outgoing_iface,
            )
        )
        log.debug(reply)
        if reply.processing_error:
            log.error(f"Error adding route: {reply.processing_error}")
            return False
        return True
     
    def remove_linux_route(
        self,
        dest_net: str,
        next_hop_addr: str,
        outgoing_iface: str,
        container: str = "",
    ) -> bool:
        """
        Remove a Linux route.
        
        Args:
            dest_net: Destination network (CIDR notation)
            next_hop_addr: Next hop IP address
            outgoing_iface: Outgoing interface name
            container: Docker container name (empty for default namespace)
            
        Returns:
            bool: True if route removed successfully, False otherwise
        """
        try:
            reply = self.client.remove_configuration(
                models.LinuxRouteConfigurationItem(
                    docker_container_name=container,
                    destination_network=dest_net,
                    next_hop_address=next_hop_addr,
                    outgoing_interface=outgoing_iface,
                )
            )
            
            errors = self._extract_reply_errors(reply)
            if errors:
                for error in errors:
                    log.error(f"Error removing Linux route: {error}")
                return False
            
            log.info(f"Linux route removed: {dest_net} via {next_hop_addr} dev {outgoing_iface}")
            return True
            
        except Exception as e:
            log.error(f"Exception removing Linux route: {e}")
            return False
            
    def add_docker_route(
        self,
        container: str,
        dest_net: str,
        next_hop_addr: str,
        outgoing_iface: str,
    ) -> bool:
        """
        Add a route to a Docker container.
        
        Args:
            container: Docker container name
            dest_net: Destination network (CIDR notation)
            next_hop_addr: Next hop IP address
            outgoing_iface: Outgoing interface name
            
        Returns:
            bool: True if route added successfully, False otherwise
        """
        return self.add_linux_route(dest_net, next_hop_addr, outgoing_iface, container)
    
    def remove_docker_route(
        self,
        container: str,
        dest_net: str,
        next_hop_addr: str,
        outgoing_iface: str,
    ) -> bool:
        """
        Remove a route from a Docker container.
        
        Args:
            container: Docker container name
            dest_net: Destination network (CIDR notation)
            next_hop_addr: Next hop IP address
            outgoing_iface: Outgoing interface name
            
        Returns:
            bool: True if route removed successfully, False otherwise
        """
        return self.remove_linux_route(dest_net, next_hop_addr, outgoing_iface, container)
        
    def add_vpp_route(
        self,
        dest_net: str,
        next_hop_addr: str,
        outgoing_iface: str,
        weight: int = 1,
    ) -> bool:
        """
        Add a VPP route.
        
        Args:
            dest_net: Destination network (CIDR notation)
            next_hop_addr: Next hop IP address
            outgoing_iface: Outgoing interface name
            weight: Route weight/priority (default: 1)
            
        Returns:
            bool: True if route added successfully, False otherwise
        """
        try:
            reply = self.client.add_configuration(
                models.RouteConfigurationItem(
                    destination_network=dest_net,
                    next_hop_address=next_hop_addr,
                    outgoing_interface=outgoing_iface,
                    weight=weight,
                )
            )
            
            errors = self._extract_reply_errors(reply)
            if errors:
                for error in errors:
                    log.error(f"Error adding VPP route: {error}")
                return False
            
            log.info(f"VPP route added: {dest_net} via {next_hop_addr} dev {outgoing_iface} weight {weight}")
            return True
            
        except Exception as e:
            log.error(f"Exception adding VPP route: {e}")
            return False
            
    def remove_vpp_route(
        self,
        dest_net: str,
        next_hop_addr: str,
        outgoing_iface: str,
        weight: int = 1,
    ) -> bool:
        """
        Remove a VPP route.
        
        Args:
            dest_net: Destination network (CIDR notation)
            next_hop_addr: Next hop IP address
            outgoing_iface: Outgoing interface name
            weight: Route weight/priority (default: 1)
            
        Returns:
            bool: True if route removed successfully, False otherwise
        """
        try:
            reply = self.client.remove_configuration(
                models.RouteConfigurationItem(
                    destination_network=dest_net,
                    next_hop_address=next_hop_addr,
                    outgoing_interface=outgoing_iface,
                    weight=weight,
                )
            )
            
            errors = self._extract_reply_errors(reply)
            if errors:
                for error in errors:
                    log.error(f"Error removing VPP route: {error}")
                return False
            
            log.info(f"VPP route removed: {dest_net} via {next_hop_addr} dev {outgoing_iface} weight {weight}")
            return True
            
        except Exception as e:
            log.error(f"Exception removing VPP route: {e}")
            return False
    
    def get_all_routes(self, container: str = "") -> List[Dict[str, Any]]:
        """
        Get all routes from a specific environment.
        
        Args:
            container: Docker container name (empty for Linux default namespace)
            
        Returns:
            List[Dict[str, Any]]: List of routes as dictionaries
        """
        try:
            # This is a placeholder implementation
            # In a real implementation, you would query the system or VPP API
            # to retrieve the actual routes
            if not container:
                # Get Linux routes
                pass
            else:
                # Get Docker container routes
                pass
            
            return []
            
        except Exception as e:
            log.error(f"Exception getting routes: {e}")
            return []

    # Note: this does not include RR-node wireguard tunnel (for that see makefile) and RR setup => only path from linux WG
    # on node to BGP in container on node
    def add_routes_for_bidirectional_bgp_communication(
        self,
        local_bgp_loop_ip: str,
        rr_bgp_loop_ip: str,
        rr_wg_ip: str,
        frr_container_name: str = "frr"
    ):
        # Some path explanation:
        # RR-to-node path: node-linux-wg0 -(adding here route1+iptables forwarding)-> FRR container default interface ->
        #   FRR container routing to local loopback -(ping echo going back, need to add route2)-> FRR container default
        #   interface -(routing path was added as part of wg configuration)-> node-linux-wg0
        # Node-to-RR path: FRR instance in FRR container -(need to add route3)-> FRR container default interface
        #   -> node linux host route(actually controlled by WG allowed ip configuration) -> wg tunnel to RR
        host_ip_from_docker_container = "172.17.0.1"

        # adding route1
        frr_container_ip_address, _ = run_command(
            [
                "docker",
                "inspect",
                "-f",
                "'{{range .NetworkSettings.Networks}}{{.IPAddress}}{{end}}'",
                frr_container_name,
            ]
        )
        frr_container_ip_address = frr_container_ip_address.replace("\n", "").replace(
            "'", ""
        )
        log.debug(f"FRR container IP address: {frr_container_ip_address}")

        # # Ensure the route does not already exist
        # existing_routes = self.get_all_routes(frr_container_name)
        # for route in existing_routes:
        #     if (
        #         route.get("destination_network") == local_bgp_loop_ip
        #         and route.get("next_hop_address") == frr_container_ip_address
        #     ):
        #         log.info(f"Route already exists: {local_bgp_loop_ip} via {frr_container_ip_address}")
        #         return
        # log.info(f"Adding route for BGP communication: {local_bgp_loop_ip} via {frr_container_ip_address}")
        # # Add the route to the FRR container

        self.add_linux_route(
            local_bgp_loop_ip,
            frr_container_ip_address,
        )
        run_command(
            [
                "sudo",
                "iptables",
                "-A",
                "FORWARD",
                "-i",
                "wg0",  # static for default wireguard interface between controller and node 
                "-o",
                "docker0",  # static for default docker installation
                "-j",
                "ACCEPT",
            ]
        )
        log.info(
            f"Added iptables rule to allow forwarding from wg0 to docker0 for FRR container {frr_container_name}"
        )
        # adding route2
        self.add_linux_route(
            rr_wg_ip,
            host_ip_from_docker_container,
            "",
            frr_container_name,
        )
        log.info(f"Added route to RR wireguard IP {rr_wg_ip} via {host_ip_from_docker_container} in FRR container {frr_container_name}")
        # adding route3
        self.add_linux_route(
            rr_bgp_loop_ip,
            host_ip_from_docker_container,
            "",
            frr_container_name,
        )
        log.info(f"Added route to RR BGP loop IP {rr_bgp_loop_ip} via {host_ip_from_docker_container} in FRR container {frr_container_name}")

    def add_ebgp_route_in_frr_container(self, client_loop_ip: str, outgoing_interface: str, frr_container_name: str = "frr"):
        """
        Add a route in the FRR container for eBGP communication.
        
        Args:
            client_loop_ip: Client loopback IP address
            frr_container_name: Name of the FRR container (default: 'frr')
        """
        # Ensure the route does not already exist
        # existing_routes = self.get_all_routes(frr_container_name)
        # for route in existing_routes:
        #     if (
        #         route.get("destination_network") == client_loop_ip
        #         and route.get("next_hop_address") == ""
        #     ):
        #         log.info(f"Route already exists: {client_loop_ip}")
        #         return
        
        log.info(f"Adding eBGP route for client loop IP: {client_loop_ip} in FRR container {frr_container_name}")
        if self.add_linux_route(client_loop_ip, "", outgoing_interface, frr_container_name):
            log.info(f"Successfully added eBGP route for client loop IP: {client_loop_ip} in FRR container {frr_container_name}")
        else:
            log.error(f"Failed to add eBGP route for client loop IP: {client_loop_ip} in FRR container {frr_container_name}")
            return False
    def add_ebgp_routes_in_vpp(
        self,
        client_loop_ip: str,
        client_interface_ip: str,
        outgoing_interface: str,
    ) -> bool:
        """
        Add eBGP routes in VPP for communication with the client.
        
        Args:
            client_loop_ip: Client loopback IP address
            client_ip: Client IP address
            outgoing_interface: Outgoing interface name
            
        Returns:
            bool: True if routes added successfully, False otherwise
        """
        try:
            # Add route to client loopback IP
            client_loop_ip = f"{client_loop_ip}/32"
            if not self.add_vpp_route(client_loop_ip, client_interface_ip, outgoing_interface):
                return False

            log.info(f"Added eBGP routes for client {client_loop_ip} via {outgoing_interface}")
            return True
            
        except Exception as e:
            log.error(f"Exception adding eBGP routes in VPP: {e}")
            return False
    
    


import subprocess


def generate_wireguard_keys():
    """
    Generate a WireGuard private & public key
    Requires that the 'wg' command is available on PATH
    Returns (private_key, public_key), both strings
    """
    privkey = subprocess.check_output("wg genkey", shell=True).decode("utf-8").strip()
    pubkey = (
        subprocess.check_output(f"echo '{privkey}' | wg pubkey", shell=True)
        .decode("utf-8")
        .strip()
    )
    return (privkey, pubkey)


wg_count_per_connection = 4
global_counter = 1
counter_per_node = {1: 0, 2: 0, 3: 0, 4: 0}

for _, (src, dst) in enumerate([(1, 2), (1, 3), (1, 4), (2, 3), (2, 4), (3, 4)]):
    # print(f"Handling WG links from VPP{src}->VPP{dst}")
    for link_index in range(2):
        if link_index == 1 and not (src == 1 and dst == 2):
            continue  # make 2 link only between node1 and node2
        for wg_group_index in range(wg_count_per_connection):
            # print(f"Handling {wg_group_index+1}.WG in WG group")
            private_key1, public_key1 = generate_wireguard_keys()
            private_key2, public_key2 = generate_wireguard_keys()
            # print(private_key1, public_key1)
            print(
                f"""
# VPP{src}<->VPP{dst}, {link_index+1}.link, {wg_group_index+1}.wg
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_NAME=wg{counter_per_node[src]}
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP=192.168.{100+global_counter}.1
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP_MASK=24
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP_NETWORK=192.168.{100+global_counter}.0/24
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_PORT={50000+global_counter}
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_PUBLIC_KEY={public_key1}
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_PRIVATE_KEY={private_key1}
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_NAME=gre{counter_per_node[src]}
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_IP_ADDRESS=192.168.{200+global_counter}.1
VPP{src}_TO_VPP{dst}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_IP_ADDRESS_MASK=24

VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_NAME=wg{counter_per_node[dst]}
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP=192.168.{100+global_counter}.2
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP_MASK=24
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_INTERFACE_IP_NETWORK=192.168.{100+global_counter}.0/24
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_PORT={50000+global_counter}
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_PUBLIC_KEY={public_key2}
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_PRIVATE_KEY={private_key2}
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_NAME=gre{counter_per_node[dst]}
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_IP_ADDRESS=192.168.{200+global_counter}.2
VPP{dst}_TO_VPP{src}_WG{link_index * wg_count_per_connection + wg_group_index}_GRE_INTERFACE_IP_ADDRESS_MASK=24"""
            )
            global_counter += 1
            counter_per_node[src] += 1
            counter_per_node[dst] += 1
